import multiprocessing
import logging
import math
import random
import parmap
import itertools
from itertools import repeat
from scipy.spatial import distance
import operator
from tqdm import tqdm
from functools import reduce
from sklearn import mixture
import statsmodels.stats.multitest as multi
import networkx as nx
import multiprocessing as mp
import numpy as np
from scipy.stats import poisson
import pandas as pd
import pygco as pygco # cut_from_graph # pip install git+git://github.com/amueller/gco_python
import scipy.stats.mstats as ms
from itertools import repeat
from scipy.stats import norm
from scipy.stats import binom
from scipy.stats import skewnorm
from scipy.sparse import issparse
import matplotlib as mpl
import matplotlib
import matplotlib.cm as cm
import matplotlib.pyplot as plt
import matplotlib.path as mplPath
import shapely.geometry
import shapely.ops
from scipy.spatial import Voronoi, voronoi_plot_2d, Delaunay, KDTree, ConvexHull
from matplotlib.patches import Polygon
from matplotlib.collections import LineCollection, PatchCollection
from PIL import Image
from matplotlib.backends.backend_pdf import PdfPages
import sklearn.manifold as manifold
import sklearn.decomposition as decomposition 
from sklearn.preprocessing import StandardScaler
from scipy.stats import gaussian_kde
from sklearn.cluster import DBSCAN,KMeans
from scipy.stats import ttest_ind
from sklearn.utils import shuffle
import hdbscan
import seaborn as sns
import sklearn.cluster as cluster
from sklearn.cluster import KMeans
from scipy.spatial.distance import cdist
from scipy import stats
import sys
import os
import time
from scipy.stats import skewnorm
from scipy import stats
from ast import literal_eval
from .Preprocessing import *
from .Graph_cut import *
from .Visualization import *


def check_target_status_jaccard(p, node, cand_seed_mat, cand_seed_size, hamming_cutoff = 0.6):   ### cutoff 0.5-0.7
#geneID = 'Ccnl1'
#newLabels=result_df.loc[geneID,][4:].values.astype(int)
#p=result_df.loc[geneID,'p_value']
#node=result_df.loc[geneID,'nodes']
#gmm = gmmDict[geneID]
#hamming_cutoff = 0.6
#if True:
    target_mat_list = list()
    for p_index in np.where(np.array(p) < 0.01)[0]:
        temp_vec = np.zeros(cand_seed_mat.shape[1])
        temp_vec[node[p_index]] = 1  
        if len(node[p_index]) <= len(temp_vec)/2:
            target_mat_list.append(temp_vec) 
 
    if len(target_mat_list) > 0:
        target_mat = np.array(target_mat_list)
        overlap_hdist = cdist(cand_seed_mat, target_mat, jaccard_dist)
#        print(np.min(overlap_hdist, axis=1))
        hit_pos = np.where(np.min(overlap_hdist, axis=1) < hamming_cutoff)[0]
        unique_hit, count_hit = np.unique(np.array(cand_seed_size)[hit_pos], return_counts = True)
        if len(target_mat_list) >= 3:
            if sum(count_hit >= len(target_mat_list)-1) >  0:
#            print('good')
                return True
            else:
                return False
        else:
            if sum(count_hit >= len(target_mat_list)-1) >  0:
                return True
            else:
                return False      
    else:
        return False

def compute_diff_vs_common_const10(u, v):
    return len(np.where((u+v) == 1)[0])/(2*len(np.where((u+v) == 2)[0]) + 10)

def jaccard_dist(u, v):
    return  1 - len(np.where((u+v) == 2)[0])/(len(np.where((u+v) > 0)[0]))


def generate_target_shape_mat(data_norm, result_df, fdr_opt):
    
    '''
    Generating target mat from these genes that needed comparing pattern.
    
    ''' 
    
    target_gene_list = list()
    target_mat_list = list()
    missed_gene = list() # save genes that are not in data_norm; and just extract from origianl data
    for geneID in fdr_opt.index.values:
        if geneID in result_df.index:
            p = result_df.loc[geneID].p_value
            node = result_df.loc[geneID].nodes
            target_gene_list.append(geneID)
            target_list = list()
            for p_index in np.arange(len(p)): 
#                if len(node[p_index]) < fdr_opt.shape[1]/2:
                temp_vec = np.zeros(data_norm.shape[0])
                temp_vec[node[p_index]] = 1  
                target_list.append(temp_vec)    
            target_mat = np.asarray(target_list)      
            target_mat_list.append(target_mat)
        else:
            missed_gene.append(geneID)
    target_df = pd.DataFrame([target_gene_list, target_mat_list]).T        
    target_df.columns = ['geneID', 'mat']
    target_df.index =  target_df.geneID     
        
    return target_df, missed_gene


def compute_p_reverLabels(locs,newLabels,gmm,exp,cellGraph):
    '''
    compute p values for each subgraph, and reverse labels of bad subgraph.
    return new reversal labels.
    '''
    
    com_factor = 1
    p_values = list()
    node_lists = list()
    reverLabel=newLabels.copy()
    a= exp[exp > 0]
    gmm_pred = gmm.predict(a.reshape(-1,1))
    zero_label = gmm.predict(np.min(a).reshape(-1,1))[0]
    label_pred = gmm.predict(exp.reshape(-1,1))
  
    if len(np.where(exp == 0)[0]) > 0:
        np.place(label_pred, exp==0, zero_label)
    gmm_pred =label_pred
    
    unique, counts = np.unique(gmm_pred,return_counts=True)
    con_components = count_component(locs,cellGraph, newLabels)
    
    for j in np.arange(len(con_components)):  ## all subgraphs are normal,score all
        if len(con_components[j]) <= 9: # len(exp)/2+10:  ## too big pattern is not reverse labels
            node_list = con_components[j]
            com_size = len(node_list)
            temp_exp = exp[np.array(list(node_list))]
            gmm_pred_com = gmm.predict(temp_exp.reshape(-1,1))
            if sum(temp_exp == 0) > 0:
#                 a=temp_exp[temp_exp>0]
#                 zero_label = gmm.predict(np.min(a).reshape(-1,1))[0]
                np.place(gmm_pred_com, temp_exp==0, zero_label)           

    # check 0s
            unique_com, counts_com = np.unique(gmm_pred_com, return_counts=True)
            major_label = unique_com[np.where(counts_com == counts_com.max())[0][0]]

            label_count = counts[np.where(unique == major_label)[0]]
            count_in_com =  counts_com.max()
            cover = exp.shape[0]/com_size

            pmf= poisson.pmf(count_in_com, com_size*(label_count/exp.shape[0]))[0]
            psf= poisson.sf(count_in_com, com_size*(label_count/exp.shape[0]))[0]
            prob_sf= (pmf+psf)*cover
            prob = (1 - poisson.cdf(count_in_com, com_size*(label_count/exp.shape[0]))[0])*cover
    #         if prob>=1:
    #             prob=1
          #  p_values.append(prob)
            #print(prob_sf,prob,major_label,com_size,cover)
          #  node_lists.append(np.array(list(node_list)))
#             for i in node_list:
#                 plt.text(locs[i,0],locs[i,1],str(i))
#             plot_voronoi_boundary(geneID,locs,exp,newLabels,prob)

            ## 1. find connected pattern of noise pattern
            if prob_sf >0.1:  
                test_node=node_list
                node_otherPattern=[]
                for i in test_node:
                    if i in cellGraph[:,0:2]:
                        node = cellGraph[np.where(i==cellGraph[:,0:2])[0],0:2]
                        node_array= node.reshape(1,-1)[0]
                        other_node= [int(val) for val in list(set(node_array)-set((test_node)))] 
                        node_otherPattern.extend(other_node)
                   #     print(i, ' ',node_array,other_node)

                node_otherPattern=np.unique(node_otherPattern)
              #  print(test_node,node_otherPattern) 
                ## 2. judege nodes of connencted other pattern in one same pattern
                diff_con=[]
                for con in range(len(con_components)):
                    if con !=j:
                        diff=set(node_otherPattern)-set(con_components[con])     
                        if len(diff)==0:
                            diff_con.append(con)
                if len(diff_con)==1:
                        #reverLabel=newLabels.copy()
                        rever =newLabels[list(con_components[diff_con[0]])][0]
                        reverLabel[list(test_node)]=rever
                    #    plot_voronoi_boundary(geneID,locs,exp,reverLabel,prob)
    return reverLabel

### 20190930 distance

def calc_distance_df( locs, data_norm,cellGraph, calc_df, tissue_mat_new, sort_tissue=True):
    
    hamming=list()
    jaccard = list()
    hausdorff = list()
    
    recal_genes=[]
    genelist=[]
    for geneID in calc_df.index:
        exp =  data_norm.loc[:,geneID].values 
        target_seed_mat_list = list()
   
        temp_factor = int(calc_df.loc[geneID].smooth_factor)  
        newLabels= calc_df.loc[geneID][5:].values.astype(int)
        node= calc_df.loc[geneID,'nodes']
        p= calc_df.loc[geneID,'p_value']
        model_labels = np.array(calc_df.loc[geneID, 'model_labels'])
        
        # num_isolate = count_isolate(locs,cellGraph, newLabels)
        # num_noise=int((num_isolate[0]+num_isolate[1])*1+(num_isolate[2]+num_isolate[3])*2+(num_isolate[4]+num_isolate[5])*3)
       # print(num_noise,len(node[0]),(locs.shape[0]-num_noise),p)

               
        if len(p)==1 and len(node[0])==locs.shape[0]:
            recal_genes.append(geneID)  ## no cuts genes, can't calculate hausdorff
                  
#         elif len(p)==1 and min(p)>0.1 and len(node[0])==(locs.shape[0]-num_noise):  ## all noise,like    
#             genelist.append(geneID)
#             total_tissue_mat=np.sum(tissue_mat_new,axis=0)
#             total_target_mat=newLabels
            
#             hamming.append(compute_diff_vs_common_const10(total_tissue_mat , total_target_mat))
#             jaccard.append(jaccard_dist(total_tissue_mat , total_target_mat))   
#             target_locs = locs[total_target_mat == 1]
#             tissue_locs = locs[total_tissue_mat == 1]
#             hausdorff.append(compute_hausdorff(target_locs,tissue_locs))
        
        else: 
            genelist.append(geneID)

    # get major label in min(p) segment
            target_seed_mat_list=list()
            node_list = node[np.argmin(p)]
            temp_vec = np.zeros(len(exp))
            temp_vec[node_list] = 1  
            target_seed_mat_list.append(temp_vec)
            # for nn in node[np.argmin(p)]:
            #     plt.text(locs[nn,0],locs[nn,1],str(nn)) 
            # plot_voronoi_boundary(geneID, locs,exp,newLabels,min(p))
            
            com_size = len(node_list)
            temp_com = np.array(list(node_list))
            pred_com = model_labels[temp_com]  
            unique_com, counts_com = np.unique(pred_com, return_counts=True)
            major_label = unique_com[np.where(counts_com == counts_com.max())[0][0]]

            max_size = 0
            size=[len(node[nn]) for nn in range(len(node))]
            p_sort=[p[i] for i in np.argsort(size)]
            node_sort=[node[i] for i in np.argsort(size)]
            #p_sort
            for p_index in np.arange(len(p)): #np.where(np.array(p) < 0.01)[0]:
                if p_sort[p_index]>min(p):  #p[p_index] <= p_cutoff and 
                    temp_vec = np.zeros(len(exp))
                    node_list = node_sort[p_index]
                    temp_com = np.array(list(node_list))
                    pred_com = model_labels[temp_com]
                    unique_com, counts_com = np.unique(pred_com, return_counts=True)
                    temp_major_label = unique_com[np.where(counts_com == counts_com.max())[0][0]]
               #     print(temp_major_label)
                    if temp_major_label == major_label :  ## get other pattern with major label
                          #['Psap'] in Rep8
                        if len(node_list)+sum(np.sum(target_seed_mat_list,axis=0))< (2/3)*len(exp)+10:
                            temp_vec[node_list] = 1  
                            target_seed_mat_list.append(temp_vec)                      
                            if len(node_list) > max_size:
                                max_size = len(node_list)
                              #  print(max_size)
                        else:
                            if sum(target_seed_mat_list[0])>max_size:
                                max_size = sum(target_seed_mat_list[0])
                    else:
                        if sum(target_seed_mat_list[0])>max_size:
                            max_size = sum(target_seed_mat_list[0])  ## sum(np.sum(target_seed_mat_list,axis=0))
                else:
                    if sum(target_seed_mat_list[0])>max_size:
                        max_size = sum(target_seed_mat_list[0])

                temp_list = list()
                if max_size > 5:
                    for nn in target_seed_mat_list:
                        if sum(nn==1) > 5 :
                            temp_list.append(nn)
                else:
                    for nn in target_seed_mat_list:
                        if sum(nn==1) == max_size:
                            temp_list.append(nn)         

            target_mat = np.array(temp_list)  
            total_target_mat = np.sum(target_mat, axis=0)  


        ## effect target mat: merge target mat as same label with min(p)   
            if sum(total_target_mat) > len(exp)/2+10:
                total_target_mat = abs(1 - total_target_mat)  ##  1-total_target_mat, for 'Hk3'


        # now find all tissue mat that with 50% of the tissue mat seg are in target
            temp_tissue_seg_list = list()
            best_overlap = list()
            best_overlap_val = 0
            match_index=[]
            for ts_index in np.arange(len(tissue_mat_new)):
            #             overlap = max(sum((tissue_mat_new[kk,:] + total_target_mat) == 2)/sum(tissue_mat_new[kk,:]), 
            #                      sum((tissue_mat_new[kk,:] + total_target_mat) == 2)/sum(total_target_mat))        
                overlap= compute_inclusion_min(tissue_mat_new[ts_index],total_target_mat)
              #  print(ts_index, overlap)
                if overlap >0.45:
                    match_index.append(ts_index)
                    temp_tissue_seg_list.append(tissue_mat_new[ts_index])


                if overlap >= best_overlap_val : #and best_overlap_val>=0.1 :
                  #  match_index.append(ts_index)
                    best_overlap_val = overlap
                    best_overlap=tissue_mat_new[ts_index].reshape(1,-1)

            if len(temp_tissue_seg_list) >0:    
                if sort_tissue==True:    
                    if len(temp_tissue_seg_list)==1 and match_index[0] > 3:
                        if match_index[0]==4 or match_index[0]==6:
                            temp_tissue_seg_list.append(tissue_mat_new[5])
                        if match_index[0]==5:
                            temp_tissue_seg_list.append(tissue_mat_new[4])
                            temp_tissue_seg_list.append(tissue_mat_new[6])
                        match_tissue_mat=np.array(temp_tissue_seg_list)
                    else:
                        match_tissue_mat = np.array(temp_tissue_seg_list)
                else:
                    match_tissue_mat = np.array(temp_tissue_seg_list)
            else:
                if best_overlap_val>0:
                    match_tissue_mat = np.array(best_overlap)
            ## 2. some points are missed by tissue mat, using min(hausdorff) to match tissue mat.['Postn'] in Rep8
                else:             ## all(overlap==0)
                    dist_temp=[]                                    
                    temp_vec=np.zeros(locs.shape[0])
                    temp_vec[:]=1
                    temp_vec[node[0]]=0  ## mostly nodes
                    u=locs[np.where(temp_vec==1)]
                    for ts_index in range(len(tissue_mat_new)):
            #                             ja_dist=jaccard_dist(temp_vec , tissue_mat_new[ts_index]) ## single noise not in any tissue mat.
            #                             ja_dist_temp.append(ja_dist)

                        v=locs[np.where(tissue_mat_new[ts_index]==1)]
                        dist=compute_hausdorff(u,v)
                        dist_temp.append(dist)
                        temp_index=np.argmin(dist_temp)
                        match_tissue_mat=tissue_mat_new[temp_index].reshape(1,-1)


            total_tissue_mat = np.sum(match_tissue_mat, axis=0) ## get tissue_mat that matched target 
            
            hamming.append(compute_diff_vs_common_const10(total_tissue_mat , total_target_mat))
            jaccard.append(jaccard_dist(total_tissue_mat , total_target_mat))   

            target_locs = locs[total_target_mat == 1]
            tissue_locs = locs[total_tissue_mat == 1]
            hausdorff.append(compute_hausdorff(target_locs,tissue_locs))
    #         print(geneID, len(target_seed_mat_list), sum(total_target_mat), 
    #               len(temp_tissue_seg_list), sum(total_tissue_mat), jaccard_dist(total_tissue_mat , total_target_mat),
    #               max(directed_hausdorff(tissue_locs, target_locs)[0], 
    #                              directed_hausdorff(target_locs, tissue_locs)[0]))
    #         plot_voronoi_boundary(geneID, locs, exp,  total_target_mat, min(p))
            #plot_voronoi_boundary(geneID, locs, exp,  total_tissue_mat, min(p))   
    #         print('------------------')
    data_array = np.array((hamming, jaccard, hausdorff), dtype=float).T
    c_labels = ['Hamming', 'Jaccard', 'Hausdorff']

    dist_df = pd.DataFrame(data_array, index=genelist, 
                      columns=c_labels)
# print out the target sum mat, and matched tissue sum mat to see whether they make sense
    return dist_df,recal_genes



def compute_inclusion_min(u, v):
#    if len(np.where((u+v) == 2)[0]) == 0:
#        return len(np.where(abs(u-v) == 1)[0])
#    else:
    return len(np.where((u+v) == 2)[0])/min(sum(v),sum(u))

def compute_hausdorff(u,v):
    '''
    Compute norm hausdorff geometry distance between two pattern.
    u: target_mat
    v: tissue_mat
    
    '''
    
    dist_u_v=distance.directed_hausdorff(u,v)[0]
    dist_v_u=distance.directed_hausdorff(v,u)[0]
    
    # if norm_factor:
    #     dist= max(dist_u_v/compute_norm_factor(v),dist_v_u/compute_norm_factor(u))
    # else: 
    return max(dist_u_v,dist_v_u)


def estimate_distance_cutoff(locs,data_norm, tissue_mat, cutoff = 0.0001):
    '''
    Estimate hamming cutoff
    ''' 
    
    target_list = list()    
                       
    target_mat = np.asarray(tissue_mat)      
    # run simulation to check the p of selected cutoff
    # tissue_mat_rand = shuffle(tissue_mat.T).T
    hdist_list = list()
    jdist_list = list()
    hfdist_list= list()
    for rr in np.arange(10): 
       # tissue_mat_rand = shuffle(tissue_mat.T).T
        target_mat_rand1 = shuffle(target_mat.T).T
        target_mat_rand2 = shuffle(target_mat.T).T   
        
        hdist_rand = cdist(target_mat_rand1, target_mat_rand2, compute_diff_vs_common_const10)
        hdist_list.append(hdist_rand.flatten())
        jdist_rand = cdist(target_mat_rand1, target_mat_rand2, jaccard_dist)   
        jdist_list.append(jdist_rand.flatten())
        
        for ta1 in range(target_mat.shape[0]):
            for ta2 in range(target_mat.shape[0]):
                target_mat_rand1_locs = locs[target_mat_rand1[ta1] == 1]
                target_mat_rand2_locs = locs[target_mat_rand2[ta2] == 1]
                hfdist_rand = compute_hausdorff(target_mat_rand1_locs,target_mat_rand2_locs)
                hfdist_list.append(hfdist_rand)
    hflattened  = [val for sublist in hdist_list for val in sublist]
    ae, loce, scalee = stats.skewnorm.fit(hflattened)
    hamming_cutoff = skewnorm.ppf(cutoff, ae, loce,scalee)
                    
    jflattened  = [val for sublist in jdist_list for val in sublist]
    ae, loce, scalee = stats.skewnorm.fit(jflattened)
    jaccord_cutoff = skewnorm.ppf(cutoff, ae, loce,scalee)
    
    
    ae, loce, scalee = stats.skewnorm.fit(hfdist_list)
    hausdorff_cutoff = skewnorm.ppf(1-cutoff, ae, loce,scalee)
    
    return hamming_cutoff, jaccord_cutoff, hausdorff_cutoff


### calculate distance

def Estimate_sf_And_dist(locs,data, cellGraph, gmmDict, test_genes, tissue_mat,plot=False,
                           unary_scale_factor=100, label_cost=10, algorithm='expansion'):
    
    '''
    Relationship smooth factor with distance for estimating smooth factor.
    the function just like recal_dist_to_tissue() with tissue_mat,
    but getting best sf by using min(hamming_dist).
    Return Best_smooth_factor, related size of gmm labels' components and three distance.
    
    '''
  
    result_df_new=pd.DataFrame()
    new_hamming=[]
    new_jaccard=[]
    new_hausdorff=[]
    con_list=[]

    for geneID in test_genes: #zeor_boundGenes: #de_counts:
        #if geneID not in result_df.index:
        count =  data.loc[:,geneID].values
        gmm=gmmDict[geneID]
        Labels = gmm.predict(count.reshape(-1,1))
        zero_label=gmm.predict(np.min(count[count>0]).reshape(-1,1))
        if sum(count==0):
            np.place(Labels,count==0, zero_label)
        con= count_component(locs, cellGraph, Labels)
        con_list.append(len(con))

        hamming_dist_list=[]
        jaccard_dist_list=[]
        hausdorff_dist_list=[]
        test_df_list=pd.DataFrame()
        
        #figsize(10,8)
        for sf in np.arange(0,110,10):  # 'Cldn9' need start from sf=0 
            temp_factor=sf
            newLabels = cut_graph_general(cellGraph, count ,gmm, unary_scale_factor, 
                                           temp_factor, label_cost, algorithm)

            #print(geneID)
            reverLabels=  compute_p_reverLabels(locs,newLabels,gmm,count,cellGraph)
            p,node, com = compute_p_CSR(locs, reverLabels, gmm, count, cellGraph) 
            
           # figsize(8,6)
            #print(sf,p)
            

            labels_array = np.array(reverLabels).reshape(1, -1)
            data_array = np.array((geneID, p, min(p),temp_factor, node), dtype=object).reshape(1,-1)
            t_array = np.hstack((data_array, labels_array))
            c_labels = ['p_value', 'fdr',  'smooth_factor', 'nodes']
            for i in np.arange(labels_array.shape[1]) + 1:
                temp_label = 'label_cell_' + str(i)
                c_labels.append(temp_label)
            test_df = pd.DataFrame(t_array[:,1:], index=t_array[:,0], 
                              columns=c_labels)

            dist_df,false_genes=calc_distance_df( locs,data,cellGraph, gmmDict,test_df, tissue_mat)
            if plot==True:
                plot_voronoi_boundary(geneID,locs,count,reverLabels,min(p))
                print(sf,p,dist_df)

            hamming_dist_list.append(dist_df.iloc[:,0].values)
            jaccard_dist_list.append(dist_df.iloc[:,1].values)
            hausdorff_dist_list.append(dist_df.iloc[:,2].values)
            test_df_list=pd.concat([test_df_list,test_df])
            
        if len(jaccard_dist_list)>0:
            min_inde =np.argmin(hamming_dist_list)  ## multiple min(p) select min(hamming_dist) amony
            
            new_hamming.append(hamming_dist_list[min_inde][0])
            new_jaccard.append(jaccard_dist_list[min_inde][0])
            new_hausdorff.append(hausdorff_dist_list[min_inde][0])
            best_test_df=test_df_list.iloc[min_inde:min_inde+1]
            result_df_new=pd.concat([result_df_new,best_test_df])
       
    
    best_sf=result_df_new.smooth_factor.values
    dist_df_new = pd.DataFrame([best_sf,con_list, new_hamming, new_jaccard,new_hausdorff]).T        ### no norm dist
    dist_df_new.columns = ['smooth_factor','con_size', 'Hamming', 'Jaccard','Hausdorff']
    dist_df_new.index =  result_df_new.index
    
    return result_df_new, dist_df_new




def recalc_dist_to_tissue(locs,data_norm, cellGraph, gmmDict, test_genes, tissue_mat_new,
                           unary_scale_factor=100, label_cost=10, algorithm='expansion'):
    
    size_factor=200
    result_df_new=pd.DataFrame()
    new_hamming=[]
    new_jaccard=[]
    new_hausdorff=[]
   
    
    noise_size_estimate =9 # min(np.min(np.sum(cand_seed_mat, axis=1)), 9)  ## #noise<=9
    # print(noise_size_estimate)
    for geneID in test_genes: #zeor_boundGenes: #de_counts:
        #if geneID not in result_df.index:
        exp =  data_norm.loc[:,geneID].values
        gmm=gmmDict[geneID]
        #print(geneID,len(np.where(exp > 0)[0]))

        test_dist_list=[]
        hamming_dist_list=[]
        jaccard_dist_list=[]
        hausdorff_dist_list=[]
        p_min=[]
        test_df_list=pd.DataFrame()
       
        #noise_cutoff=np.inf
        
        temp_factor = 0   # 'Cldn9' need start from sf=0 
        newLabels = cut_graph_general(cellGraph, exp,gmm, unary_scale_factor, 
                                           temp_factor, label_cost, algorithm)

        p, node, com = compute_p_CSR(locs, newLabels, gmm, exp, cellGraph)
        num_isolate = count_isolate(locs,cellGraph, newLabels)  

        noise_size = sum(num_isolate[0:int(2*noise_size_estimate)]) 
        normalized_noise = noise_size*(size_factor/len(exp))
        noise_cutoff=normalized_noise
        
        
        labels_array = np.array(newLabels).reshape(1, -1)
        data_array = np.array((geneID, p, min(p),temp_factor, node), dtype=object).reshape(1,-1)
        t_array = np.hstack((data_array, labels_array))
        c_labels = ['p_value', 'fdr',  'smooth_factor', 'nodes']
        for i in np.arange(labels_array.shape[1]) + 1:
            temp_label = 'label_cell_' + str(i)
            c_labels.append(temp_label)
        test_df = pd.DataFrame(t_array[:,1:], index=t_array[:,0], 
                          columns=c_labels)

        dist_df,false_genes=calc_distance_df( locs, data_norm,cellGraph, gmmDict,test_df, tissue_mat_new)
     #   print(geneID,temp_factor,noise_cutoff,dist_df,false_genes)
        
        if noise_cutoff<6 and noise_cutoff>0:  ## 'Vstm4' have no 0<noise<5 , just skip 5.
#                     print(dist_df)
#                     print('---')
            hamming_dist_list.append(dist_df.iloc[:,0].values)
            jaccard_dist_list.append(dist_df.iloc[:,1].values)
            hausdorff_dist_list.append(dist_df.iloc[:,2].values)
            p_min.append(min(p))
            test_df_list=pd.concat([test_df_list,test_df])
            
        while noise_cutoff >0 and len(false_genes)==0 and temp_factor<=80:
      #  for sf_add in np.arange(10,add_sf,10):   
            
            temp_factor = temp_factor+5
            newLabels = cut_graph_general(cellGraph, exp,gmm, unary_scale_factor, 
                                               temp_factor, label_cost, algorithm)

            p, node, com = compute_p_CSR(locs, newLabels, gmm, exp, cellGraph)
            num_isolate = count_isolate(locs,cellGraph, newLabels)  
            
            noise_size = sum(num_isolate[0:int(2*noise_size_estimate)]) 
            normalized_noise = noise_size*(size_factor/len(exp))
            
            noise_cutoff=normalized_noise
           # print(geneID,temp_factor,noise_cutoff)
            #plot_voronoi_boundary(geneID,locs,exp,newLabels,min(p))
            
            labels_array = np.array(newLabels).reshape(1, -1)
            data_array = np.array((geneID, p, min(p),temp_factor, node), dtype=object).reshape(1,-1)
            t_array = np.hstack((data_array, labels_array))
            c_labels = ['p_value', 'fdr',  'smooth_factor', 'nodes']
            for i in np.arange(labels_array.shape[1]) + 1:
                temp_label = 'label_cell_' + str(i)
                c_labels.append(temp_label)
            test_df = pd.DataFrame(t_array[:,1:], index=t_array[:,0], 
                              columns=c_labels)

            dist_df,false_genes=calc_distance_df( locs, data_norm,cellGraph, gmmDict,test_df, tissue_mat_new)
            
            
            if noise_cutoff<6 and noise_cutoff>0:
#                     print(dist_df)
#                     print('---')
                hamming_dist_list.append(dist_df.iloc[:,0].values)
                jaccard_dist_list.append(dist_df.iloc[:,1].values)
                hausdorff_dist_list.append(dist_df.iloc[:,2].values)
                p_min.append(min(p))
                test_df_list=pd.concat([test_df_list,test_df])
                    
#             if len(false_genes)!=0: #or noise_cutoff==0:
#                 print('false_genes:',false_genes)
#                 break;


        if len(hausdorff_dist_list)>0:
            best_p=min(p_min)  ## multiple min(p) select min(jaccard_dist) amony
            best_p_index=np.where(np.array(p_min)==best_p)[0]
            min_p_haus=[hausdorff_dist_list[i] for i in best_p_index]
            min_inde=np.where(np.array(hausdorff_dist_list)==min(min_p_haus))[0][0]
            #min_inde=np.argmin(p_min)  #jaccard_dist_list)   ## min(p) as selection standard.
            #min_inde_haus=np.argmin(hausdorff_dist_list)
            new_hamming.append(hamming_dist_list[min_inde][0])
            new_jaccard.append(jaccard_dist_list[min_inde][0])
            new_hausdorff.append(hausdorff_dist_list[min_inde][0])
            best_test_df=test_df_list.iloc[min_inde:min_inde+1]
            result_df_new=pd.concat([result_df_new,best_test_df])

    dist_df_new = pd.DataFrame([new_hamming, new_jaccard,new_hausdorff]).T        ### no norm dist
    dist_df_new.columns = ['Hamming', 'Jaccard','Hausdorff']
    dist_df_new.index =  result_df_new.index

    return result_df_new, dist_df_new




## abundan
def computer_norm_jaccard_to_tissue(tissue_mat, target_df, fdr_opt):
    '''
    computer jaccard distance
    
    '''
    jaccard_result = list()
    temp_result_df = fdr_opt[fdr_opt.fdr < 0.01].sort_values(by=['fdr'])
    p_cutoff = min(temp_result_df.iloc[temp_result_df.shape[0]-1].p_value)
#    p_cutoff = min(fdr_opt.p_value[int(np.where(fdr_opt.fdr == max(fdr_opt.fdr))[0])])
    for geneID in target_df.index:
        p = fdr_opt.loc[geneID].p_value
        node = fdr_opt.loc[geneID].nodes
        sizes=list()
        for nn in node:
            sizes.append(len(nn))

        temp_mat = target_df.loc[geneID].mat
        overlap_hdist = cdist(tissue_mat, temp_mat, compute_inclusion_new)
    #    mapping_index_c = np.argmax(overlap_hdist, axis=0)
        mapping_index_r = np.argmax(overlap_hdist, axis=1)
        
        jaccard_val = list()
        for ts_index in np.arange(tissue_mat.shape[0]):
            match_index = np.where(overlap_hdist[ts_index] > 0.6)[0]
            # use small seg first, use same number of seg as seed:
            # until size match
            if len(match_index) > 0:
                temp_vec= np.zeros(tissue_mat.shape[1])
                for m_index in match_index:
                    if sum(temp_vec) < 1.2*sum(tissue_mat[ts_index]):
                        temp_vec = temp_vec + temp_mat[m_index]
                temp_jaccard = distance.jaccard(tissue_mat[ts_index], temp_vec)
                jaccard_val.append(temp_jaccard)         
        p = fdr_opt.loc[geneID].p_value
        node = fdr_opt.loc[geneID].nodes
        if geneID == 'Fgfr1':
            print(p_cutoff, p)
        for p_index in np.arange(len(p)):
            if p[p_index] <= p_cutoff:
                temp_vec= np.zeros(tissue_mat.shape[1])
                temp_vec[node[p_index]] = 1
                for ts_index in np.arange(tissue_mat.shape[0]):
                    temp_jaccard = distance.jaccard(tissue_mat[ts_index], temp_vec) #,compute_jaccard)
                    jaccard_val.append(temp_jaccard)  
                    if geneID == 'Fgfr1':
                        print(p_cutoff, p)
                        print(p_index, ts_index, sum(temp_vec), sum(tissue_mat[ts_index]), temp_jaccard)

                        
        if len(jaccard_val) == 0:
            ts_index = np.argmin(np.sum(tissue_mat, axis=1))
            temp_vec= np.zeros(tissue_mat.shape[1])
            one_index = random.sample(range(0, tissue_mat.shape[1]), sum(tissue_mat[ts_index]).astype(np.int))
            temp_vec[one_index] = 1
            temp_jaccard = distance.jaccard(tissue_mat[ts_index], temp_vec) #,compute_jaccard)# temp_hamming = 10 #random.uniform(2.5, 4)   #compute_diff_vs_common_new(tissue_mat[ts_index], temp_vec)
            jaccard_val.append(temp_jaccard)
        jaccard_result.append(min(jaccard_val))
    jaccard_df = pd.DataFrame([target_df.index, jaccard_result]).T        
    jaccard_df.columns = ['geneID', 'jaccard_dist']
    jaccard_df.index =  target_df.index
    return jaccard_df

def compute_inclusion_new(u, v):
#    if len(np.where((u+v) == 2)[0]) == 0:
#        return len(np.where(abs(u-v) == 1)[0])
#    else:
        return len(np.where((u+v) == 2)[0])/sum(v)


## abundan
def computer_norm_hamming_to_tissue(tissue_mat, target_df, fdr_opt):
    '''
    computer hamming distance
    '''

    hamming_result = list()
    temp_result_df = fdr_opt[fdr_opt.fdr < 0.01].sort_values(by=['fdr'])
    p_cutoff = min(temp_result_df.iloc[temp_result_df.shape[0]-1].p_value)
    for geneID in target_df.index:
        p = fdr_opt.loc[geneID].p_value
        node = fdr_opt.loc[geneID].nodes
        sizes=list()
        for nn in node:
            sizes.append(len(nn))
        temp_mat = target_df.loc[geneID].mat
        overlap_hdist = cdist(tissue_mat, temp_mat, compute_inclusion_new)
    #    mapping_index_c = np.argmax(overlap_hdist, axis=0)
        mapping_index_r = np.argmax(overlap_hdist, axis=1)
        
        hamming_val = list()
        for ts_index in np.arange(tissue_mat.shape[0]):
            match_index = np.where(overlap_hdist[ts_index] > 0.6)[0]
            # use small seg first, use same number of seg as seed:
            # until size match
            if len(match_index) > 0:
                temp_vec= np.zeros(tissue_mat.shape[1])
                for m_index in match_index:
                    if sum(temp_vec) < 1.2*sum(tissue_mat[ts_index]):
                        temp_vec = temp_vec + temp_mat[m_index]
                temp_hamming = compute_diff_vs_common_new(tissue_mat[ts_index], temp_vec)
                hamming_val.append(temp_hamming)         
        
        if geneID == 'Fgfr1':
            print(p_cutoff, p)
        for p_index in np.arange(len(p)):
            if p[p_index] <= p_cutoff:
                temp_vec= np.zeros(tissue_mat.shape[1])
                temp_vec[node[p_index]] = 1
                for ts_index in np.arange(tissue_mat.shape[0]):
                    temp_hamming = compute_diff_vs_common_new(tissue_mat[ts_index], temp_vec)
                    hamming_val.append(temp_hamming)  
                    if geneID == 'Fgfr1':
                        print(p_cutoff, p)
                        print(p_index, ts_index, sum(temp_vec), sum(tissue_mat[ts_index]), temp_hamming)

                        
        if len(hamming_val) == 0:
            ts_index = np.argmin(np.sum(tissue_mat, axis=1))
            temp_vec= np.zeros(tissue_mat.shape[1])
            one_index = random.sample(range(0, tissue_mat.shape[1]), sum(tissue_mat[ts_index]).astype(np.int))
            temp_vec[one_index] = 1
            temp_hamming = compute_diff_vs_common_new(tissue_mat[ts_index], temp_vec)# temp_hamming = 10 #random.uniform(2.5, 4)   #compute_diff_vs_common_new(tissue_mat[ts_index], temp_vec)
            hamming_val.append(temp_hamming)
        hamming_result.append(min(hamming_val))
    hamming_df = pd.DataFrame([target_df.index, hamming_result]).T        
    hamming_df.columns = ['geneID', 'hamming']
    hamming_df.index =  target_df.index
    return hamming_df

def compute_diff_vs_common_const_new(u, v):
    return len(np.where((u+v) == 1)[0])/(2*len(np.where((u+v) == 2)[0]) + 10)

def compute_diff_vs_common_new(u, v):
    return len(np.where((u+v) == 1)[0])/(2*len(np.where((u+v) == 2)[0]) + 10)



## hausdorff distance

def compute_norm_dist_to_tissue_parallel(locs, tissue_mat_new, target_df, result_df):
    
    num_cores = mp.cpu_count()
    if num_cores > math.floor(target_df.shape[0]/2):
         num_cores=int(math.floor(target_df.shape[0]/2))
    ttt = np.array_split(target_df,num_cores,axis=0)
    tuples = [(l, d, u,  c) for l, d, u, c in zip(repeat(locs, num_cores),
                                                        repeat(tissue_mat_new, num_cores),
                                                        ttt,     
                                    repeat(result_df, num_cores))]
                                   # repeat(p_cutoff, num_cores))] 
                                    
    dist_results = parmap.starmap(compute_norm_dist_to_tissue, tuples,
                             pm_processes=num_cores, pm_pbar=True)
    
    dd=[dist_results[i][0] for i in range(len(dist_results))]
    dist_val=reduce(operator.add,dd)
    
    gg=[dist_results[i][1] for i in range(len(dist_results))]
    genes=reduce(operator.add,gg)
    
    re=[dist_results[i][2] for i in range(len(dist_results))]
    recal_genes=reduce(operator.add,re)
    
    dist_norm=[val /max(dist_val) for val in dist_val]
    dist_df = pd.DataFrame([genes, dist_val,dist_norm]).T        
    dist_df.columns = ['geneID', 'dist','norm_dist']
    dist_df.index =  genes
    
    return dist_df,recal_genes

    
def compute_norm_dist_to_tissue(locs,tissue_mat_new,target_df,result_df):
    
    '''
    u: targrt vector
    v: tissue vector
    '''
    
    dist_result=[]
    recal_genes=[]
    genelist=[]
    for geneID in target_df.index:
        
        #exp=data_norm.loc[:,geneID].values
        fdr = result_df.loc[geneID].fdr
        p=result_df.loc[geneID].p_value
        node = result_df.loc[geneID].nodes
        newLabels=result_df.loc[geneID][4:].values.astype(int)
        sf=result_df.loc[geneID,'smooth_factor']
        #plot_voronoi_boundary(geneID,locs,exp,newLabels,p=fdr)

       
        ### 1. sum(exp>0)<3,just one or two noise, no in target df, comparing noise and all tissue patterns.
        ###  no exist len(p)==1 and len(node[0])<3. 
        # geneID='Gm16532'
        if len(p)==1 and len(node[0])>locs.shape[0]-3 and len(node[0])<locs.shape[0]: 
            genelist.append(geneID)
            dist_temp=[]
            temp_vec=np.zeros(locs.shape[0])
            temp_vec[:]=1
            temp_vec[node[0]]=0
            u=locs[np.where(temp_vec==1)]
            for ts_index in range(len(tissue_mat_new)):
                v=locs[np.where(tissue_mat_new[ts_index]==1)]
                dist=compute_hausdorff(u,v)
                dist_temp.append(dist)
            dist_result.append(min(dist_temp))
       # dist_result.append(dist_val)
            # zero_boundGenes.append(geneID)
            
            # temp_hull=ConvexHull(locs)
            # u=locs[temp_hull.simplices.reshape(1,-1)[0]] 

            # dist_temp=[]
            # for ts_index in np.arange(len(tissue_mat_new)):
            #     v=locs[np.where(tissue_mat_new[ts_index]==1)]
            #     norm_factor=compute_norm_factor(v)
            #     dist=distance.directed_hausdorff(u,v)[0]
            #     dist_norm=dist/norm_factor
            #     # print(ts_index,norm_factor,dist,dist_norm)
            #     if dist_norm>1:
            #         dist_norm=1
            #     dist_temp.append(dist_norm)
            # dist_val=max(dist_temp)

        ## 2. no boundarys (with big smooth factor) , recalc with a series of small sf, not in target df.
        # geneID='Map1a'
        elif len(p)==1 and len(node[0])==locs.shape[0]:
            #dist_temp=[]
            recal_genes.append(geneID)

        ## 3. normal graph cuts, each target match to tissue.                
                
        else:
            genelist.append(geneID)
            dist_temp=[]            
            temp_mat=target_df.loc[geneID,'mat']
            
            overlap_hdist=cdist(tissue_mat_new,temp_mat,compute_inclusion_min)
            
            for tg_index in np.arange(len(temp_mat)):   ## tg_index: target index; match_ts_index: tissue index
                    match_ts_index=np.where(overlap_hdist.T[tg_index].T > 0.42)[0]
                   # print(match_ts_index,tg_index)
                    ## one targrt -to -one tissue
                    if len(match_ts_index)==1:  # match one tissue
                        #print(match_ts_index,tg_index)

                        u=locs[np.where(temp_mat[tg_index]==1)]  ## target
                        v=locs[np.where(tissue_mat_new[match_ts_index[0]]==1)]  ## tissue
                    # print(distance.directed_hausdorff(u,v)[0])
                        dist=compute_hausdorff(u,v)
                       # print(dist)
                        dist_temp.append(dist)
                        
                    ## one target -to multiple tissue
            #      ## one target match multiple tissue, should merge tissue
                    elif len(match_ts_index)>=2:  
                        #print(match_ts_index,tg_index)
                        tissue_vec = np.zeros(locs.shape[0])  ## tissue merge
                        for m_index in match_ts_index:
                            tissue_vec =tissue_vec+ tissue_mat_new[m_index]

                        u=locs[np.where(temp_mat[tg_index]==1)] 
                        v=locs[np.where(tissue_vec==1)]
                        dist=compute_hausdorff(u,v)
                       # print(dist)
                        dist_temp.append(dist)
                        
                    else:    ## target match zero tissue, (don't match tissue).
                    #    print('{} pattern no match:'.format(tg_index))      
                        for ts_index in np.arange(len(tissue_mat_new)):
                            u=locs[np.where(temp_mat[tg_index]==1)] 
                            v=locs[np.where(tissue_mat_new[ts_index]==1)]
                            dist=compute_hausdorff(u,v)
                           # print(dist)
                            dist_temp.append(dist)
            dist_result.append(min(dist_temp))
#     dist_df = pd.DataFrame([target_df.index, dist_result]).T        
#     dist_df.columns = ['geneID', 'dist']
#     dist_df.index =  target_df.index
    return dist_result,genelist, recal_genes


def compute_inclusion_min(u, v):
#    if len(np.where((u+v) == 2)[0]) == 0:
#        return len(np.where(abs(u-v) == 1)[0])
#    else:
    return len(np.where((u+v) == 2)[0])/min(sum(v),sum(u))


def compute_hausdorff(u,v):
    '''
    Compute norm hausdorff geometry distance between two pattern.
    u: target_mat
    v: tissue_mat
    
    '''
    
    dist_u_v=distance.directed_hausdorff(u,v)[0]
    dist_v_u=distance.directed_hausdorff(v,u)[0]
    
    # if norm_factor:
    #     dist= max(dist_u_v/compute_norm_factor(v),dist_v_u/compute_norm_factor(u))
    # else:     
    return max(dist_u_v,dist_v_u)
    

def compute_norm_factor(coord):
    norm_dist=[]
    for i in range(coord.shape[0]):
        for j in range(coord.shape[0]):
            temp_dist = distance.euclidean(coord[i], coord[j])
            norm_dist.append(temp_dist)
    return max(norm_dist)


def compute_norm_hausdorff(u,v,norm_factor=True):
    '''
    Compute norm hausdorff geometry distance between two pattern.
    u: target_mat
    v: tissue_mat
    norm_factor: the number of pattern in positive tissue
    '''
    
    dist_u_v=distance.directed_hausdorff(u,v)[0]
    dist_v_u=distance.directed_hausdorff(v,u)[0]
    
    if norm_factor:
        dist= max(dist_u_v/compute_norm_factor(v),dist_v_u/compute_norm_factor(u))
    else:
        dist= max(dist_u_v,dist_v_u)
        
    return dist


def simulate_distance_cutoff(locs,data_norm, tissue_mat, cutoff = 0.0001):
    '''
    Estimate hamming cutoff
    ''' 
    
    target_list = list()    
                       
    target_mat = np.asarray(tissue_mat)      
    # run simulation to check the p of selected cutoff
    # tissue_mat_rand = shuffle(tissue_mat.T).T
    hdist_list = list()
    jdist_list = list()
    hfdist_list= list()
    for rr in np.arange(10): 
       # tissue_mat_rand = shuffle(tissue_mat.T).T
        target_mat_rand1 = shuffle(target_mat.T).T
        target_mat_rand2 = shuffle(target_mat.T).T   
        
        hdist_rand = cdist(target_mat_rand1, target_mat_rand2, compute_diff_vs_common_const10)
        hdist_list.append(hdist_rand.flatten())
        jdist_rand = cdist(target_mat_rand1, target_mat_rand2, jaccard_dist)   
        jdist_list.append(jdist_rand.flatten())
        
        for ta1 in range(target_mat.shape[0]):
            for ta2 in range(target_mat.shape[0]):
                target_mat_rand1_locs = locs[target_mat_rand1[ta1] == 1]
                target_mat_rand2_locs = locs[target_mat_rand2[ta2] == 1]
                hfdist_rand = compute_hausdorff(target_mat_rand1_locs,target_mat_rand2_locs)
                hfdist_list.append(hfdist_rand)
    hflattened  = [val for sublist in hdist_list for val in sublist]
    ae, loce, scalee = stats.skewnorm.fit(hflattened)
    hamming_cutoff = skewnorm.ppf(cutoff, ae, loce,scalee)
                    
    jflattened  = [val for sublist in jdist_list for val in sublist]
    ae, loce, scalee = stats.skewnorm.fit(jflattened)
    jaccord_cutoff = skewnorm.ppf(cutoff, ae, loce,scalee)
    
    
    ae, loce, scalee = stats.skewnorm.fit(hfdist_list)
    hausdorff_cutoff = skewnorm.ppf(1-cutoff, ae, loce,scalee)
    
    return hamming_cutoff, jaccord_cutoff, hausdorff_cutoff


def create_tissue_mat_new(locs,cellGraph,tissue_mat):
    '''
    devide tissue mat into single connect patterns
    '''
    
    effect_ind=[]
    
    for ts in range(len(tissue_mat)):
        temp_tissue=tissue_mat[ts]
        com=count_component(locs,cellGraph,temp_tissue)
        connectInd=[i for i in range(len(com)) if len(com[i])>=5 and len(com[i])<tissue_mat.shape[1]/4]
        #print(ts,connectInd)
        if len(connectInd)==2 or len(connectInd)==3:
            print(ts)
            effect_ind.append(ts)
            
    temp_tissue_other=np.zeros(tissue_mat.shape[1])
    for ee in effect_ind:
        temp_tissue_other=temp_tissue_other+tissue_mat[ee]
    #
    plt.scatter(locs[:,0],locs[:,1],c=temp_tissue_other)  
    #plt.show()
    com=count_component(locs,cellGraph,temp_tissue_other)
    connectInd=[i for i in range(len(com)) if len(com[i])>=3 and len(com[i])<tissue_mat.shape[1]/2]
    #connectInd
    tissue_mat_new=list()
    for i in connectInd:
        temp_graph=np.zeros(tissue_mat.shape[1])
        temp_graph[list(com[i])]=1
        plt.scatter(locs[:,0],locs[:,1],c=temp_graph)

        print('-----com_rank{}---com_size{}'.format(i,len(com[i])))
        tissue_mat_new.append(temp_graph)
        plt.show()
        
    return tissue_mat_new



## recalculate no cuts genes
def recalc_dist_to_tissue_parallel(locs,data_norm, cellGraph, gmmDict, test_genes,tissue_mat_new, add_sf=30,
                           unary_scale_factor=100, label_cost=10, algorithm='expansion'):
    num_cores = mp.cpu_count()
    if num_cores > math.floor(len(test_genes)/2):
         num_cores=int(math.floor(len(test_genes)/2))
    ttt = np.array_split(test_genes,num_cores,axis=0)
    tuples = [(l, d,c,g,t,ts, a, u, ll, al) for l, d,c,g,t,ts, a, u, ll, al in zip(repeat(locs, num_cores),
                                                        repeat(data_norm, num_cores),
                                                        repeat(cellGraph, num_cores),
                                                        repeat(gmmDict, num_cores),                             
                                                        ttt, 
                                                        repeat(tissue_mat_new, num_cores),
                                                        repeat(add_sf, num_cores),
                                                        repeat(unary_scale_factor, num_cores),
                                                        repeat(label_cost, num_cores),
                                                        repeat(algorithm, num_cores))]
    dist_results = parmap.starmap(recalc_dist_to_tissue, tuples,
                             pm_processes=num_cores, pm_pbar=True)
    
    result_df_new=pd.DataFrame()
    best_dist_df=pd.DataFrame()
    for i in range(len(dist_results)):
        result_df_new=pd.concat([result_df_new,dist_results[i][0]])
        best_dist_df=pd.concat([best_dist_df,dist_results[i][1]])
    
    return result_df_new,best_dist_df

def cumcdf(data,num_bin=300):
    counts,bin_edges=np.histogram(data,bins=num_bin)
    cdf=np.cumsum(counts)
    x=bin_edges[1:]
    y=cdf/cdf[-1]
    return x,y


import numpy as np
import pandas as pd
from sklearn import mixture
import statsmodels.stats.multitest as multi
import multiprocessing as mp
import math
import parmap
from itertools import repeat

def find_mixture(data, random_state = 2020):
    '''
    estimate expression clusters
    
    :param file: data (n,); 
    :rtype: gmm object;
    '''

    #n_components_range = range(2,5)
    best_component = 2
    lowest_bic=np.infty

    data = np.unique(data)  # all duplications, when data[data>0]
    data = list(data).copy()

    if len(data)<=2:   # 20211119 coco
        # len(data) == 0, data=[0,0], n = 1
        # len(data) == 1, data=[x,0], n=2
        # len(data) == 2, data[x1,x2], n= 2
        # len(data) == 3/4,  , n = [2,3/4]
        # len(data) >=5, , n= range(2,5)
        zero = np.repeat([0],3-len(data))
        data.extend(zero)
        temp_data = np.array(data).reshape(-1,1)
        n_components = min(len(np.unique(data)),2)
        gmm = mixture.GaussianMixture(n_components = n_components,random_state=random_state)
        gmm.fit(temp_data)
        best_gmm=gmm
    else:
        temp_data = np.array(data).reshape(-1,1)
        if len(temp_data)<5 and len(temp_data)>2:
            n_components_range=range(2,len(temp_data))

        else:
             n_components_range = range(2,5)

        for n_components in n_components_range:
            gmm = mixture.GaussianMixture(n_components = n_components,random_state=random_state)
            gmm.fit(temp_data)
            gmm_bic = gmm.bic(temp_data)
            if gmm_bic < lowest_bic:
                best_gmm = gmm
                lowest_bic = gmm_bic
                best_component = n_components      

    return best_gmm


def find_mixture_2(data,random_state=2020):
    '''
    estimate expression clusters, use k=2
    
    :param file: data (n,); 
    :rtype: gmm object;
    '''
    data = np.unique(data)
    data = list(data).copy()

    if len(data)<=2:
        zero = np.repeat([0],3-len(data))
        data.extend(zero)
    n_components = min(len(np.unique(data)), 2)
    temp_data = np.array(data).reshape(-1,1)
    gmm = mixture.GaussianMixture(n_components = n_components,random_state=random_state)
    gmm.fit(temp_data)

    return gmm


def find_mixture_n(data,n_components = 2, random_state=2020):
    '''
    estimate expression clusters, use k=2
    
    :param file: data (n,); 
    :rtype: gmm object;
    '''
    data = np.unique(data)
    data = list(data).copy()

    if len(data)<=2:
        zero = np.repeat([0],3-len(data))
        data.extend(zero)
    n_components = min(len(np.unique(data)), n_components)
    temp_data = np.array(data).reshape(-1,1)
    gmm = mixture.GaussianMixture(n_components = n_components,random_state=random_state)
    gmm.fit(temp_data)

    return gmm



def perform_gmm(count, random_state = 2020):
    '''
    reture one trained gmm model
    
    '''
    a = count.copy()
#    if sum(a>1)/len(count) <= 0.1 and sum(a>1)<=30:
#        np.place(a, a==0, (np.random.rand(sum(a==0))*0.25)) 
#        gmm = find_mixture_2(a)       
#    else:
    if True:
        a=a[a>0]
        gmm = find_mixture(a, random_state=random_state)
        gmm_pred = gmm.predict(count.reshape(-1,1))
        unique, counts = np.unique(gmm_pred,return_counts=True)
        if np.min(counts) < 0.1*len(count):
            gmm = find_mixture_2(a, random_state= random_state) 
    return gmm


def gmm_model(data_norm, random_state = 2020):
    gmmDict_={}
    for geneID in data_norm.columns:
        count=data_norm.loc[:,geneID].values
        gmm=perform_gmm(count, random_state = random_state)   
        gmmDict_[geneID]=gmm
    return gmmDict_


def multiGMM(data_norm,random_state=2020, ncores = None):
    all_cores = mp.cpu_count()
    # ncores = input('You can select your number of cores: ')
    # print(ncores)
    if ncores !=None:
        num_cores = ncores
    else:
        num_cores = int(all_cores*0.6)
    if num_cores > math.floor(data_norm.shape[1]/2):
        num_cores=int(math.floor(data_norm.shape[1]/2))
    print(f'scGCO used {num_cores} out of {all_cores} cores')

    ttt = np.array_split(data_norm,num_cores,axis=1)
    tuples = [(d, rd) for d, rd in zip(ttt, 
                                    repeat(random_state, num_cores))] 
    gmmDict_=parmap.starmap(gmm_model, tuples,
                             pm_processes=num_cores, pm_pbar=True)
    gmmDict={}
    for i in np.arange(len(gmmDict_)):
        gmmDict.update(gmmDict_[i])   ## dict.update()  add dict to dict
    return gmmDict


def gmm_model_2(data_norm,random_state=2020 ):
    gmmDict_={}
    for geneID in data_norm.columns:
        count=data_norm.loc[:,geneID].values
        gmm=find_mixture_2(count,random_state=random_state)   
        gmmDict_[geneID]=gmm
    return gmmDict_


def multiGMM_2(data_norm,random_state=2020):
    '''
    GMM with conponents =2
    '''
    num_cores = mp.cpu_count()
    num_cores = int(num_cores*0.5)
    if num_cores > math.floor(data_norm.shape[1]/2):
        num_cores=int(math.floor(data_norm.shape[1]/2))
   # print(num_cores)
    ttt = np.array_split(data_norm,num_cores,axis=1)
    #print(ttt)

    tuples = [(d, rd) for d, rd in zip(ttt, 
                                    repeat(random_state, num_cores))] 
    gmmDict_=parmap.starmap(gmm_model, tuples,
                             pm_processes=num_cores, pm_pbar=True)
    gmmDict={}
    for i in np.arange(len(gmmDict_)):
        gmmDict.update(gmmDict_[i])   ## dict.update()  add dict to dict
    return gmmDict




def TSNE_gmm(data):
    n_components_range = range(2,10)
    best_component = 2
    lowest_bic=np.infty
    temp_data = data
    for n_components in n_components_range:
        gmm = mixture.GaussianMixture(n_components = n_components,random_state=2020)
        gmm.fit(temp_data)
        gmm_bic = gmm.bic(temp_data)
        if gmm_bic < lowest_bic:
            best_gmm = gmm
            lowest_bic = gmm_bic
            best_component = n_components     
    return best_gmm



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
import sklearn.manifold as manifold
import sklearn.decomposition as decomposition 
from sklearn.preprocessing import StandardScaler
from scipy.stats import gaussian_kde
from sklearn.cluster import DBSCAN,KMeans
import pygco
import umap
from .Graph_cut import compute_pairwise_cost


# clustering genes
def spatial_pca_UMAP_kmeans_cluster_gene(data_norm, gene_lists,marker_genes,n_neighbors=10,min_dist=0.1,
                                         plots=True, fileName=None):
    data_s = StandardScaler().fit_transform(data_norm.loc[:, gene_lists])
    pca = decomposition.PCA()
    pca.fit(data_s.T)
    pca_proj = pca.fit_transform(data_s.T)

    num_comp = np.where(np.cumsum(pca.explained_variance_)/np.sum(pca.explained_variance_)
                    > 0.90)[0][0]
    umap_proj = umap.UMAP(n_neighbors=n_neighbors,  # 这决定了流形结构局部逼近中相邻点的个数。更大的值将导致更多的全局结构被保留，而失去了详细的局部结构。一般来说，这个参数应该在5到50之间，10到15是一个合理的默认值。
                          min_dist=min_dist,  #这控制了嵌入的紧密程度，允许压缩点在一起。数值越大，嵌入点分布越均匀;数值越小，算法对局部结构的优化越精确。合理的值在0.001到0.5之间，0.1是合理的默认值。
                          n_components=2,
#                           random_state = 0
                         ).fit_transform(pca_proj[:,0:num_comp])  ## genes cluster
#     print(umap_proj.shape)

    umap_proj_df=pd.DataFrame(index=gene_lists)
    umap_proj_df["UMAP1"]=umap_proj[:,0]
    umap_proj_df["UMAP2"]=umap_proj[:,1]
    init = umap_proj_df.loc[marker_genes].values
    num_clusters=len(marker_genes)
    kmeans=KMeans(n_clusters=num_clusters,
                  init = init, 
#                   random_state=0
                 ).fit(umap_proj)

    umap_proj_df["cluster"]=kmeans.labels_
    gene_subset_lists=list()
    for geneID in marker_genes:
        gene_subset = umap_proj_df.index[np.where(umap_proj_df.cluster == umap_proj_df.loc[geneID,"cluster"])]
        gene_subset_lists.append(gene_subset)

    for i ,gene_subset in enumerate(gene_subset_lists):
        umap_proj_df.loc[gene_subset,"cluster"]=i
    
    if plots==True:
        final_labels = umap_proj_df.cluster.values
        final_tsne = np.c_[umap_proj, final_labels]
        palette = sns.color_palette('deep', final_labels.max() + 1)
        colors = [palette[x] if x >= 0 else (0.0, 0.0, 0.0) for x in final_tsne[:,2].astype(int)]
        plt.scatter(final_tsne[:,0], final_tsne[:,1], c=colors, s=28)
        plt.xlabel('UMAP component 1')
        plt.ylabel('UMAP component 2')
        for i in final_labels:
            position = np.max(final_tsne[ final_tsne[:,2]== i], axis=0)
            plt.gcf().gca().text(position[0], position[1]-1,str(i), fontsize=12)
        if fileName != None:
            plt.savefig(fileName,format="pdf",dpi=300)
        plt.show()
    return umap_proj_df



def spatial_pca_tsne_kmeans_cluster_gene(data_norm, gene_lists,marker_genes, perplexity = 30,fileName=None):

    '''
    perform standard PCA + tsne
    :param file: data_norm: normalized gene expression; gene_lists: list shape(k,)
        perplexity = 30
    :rtype: tsne_proj: shape (m, 2)
    '''          
    data_s = StandardScaler().fit_transform(data_norm.loc[:, gene_lists])
    pca = decomposition.PCA()
    pca.fit(data_s.T)
    pca_proj = pca.fit_transform(data_s.T)
    num_comp = np.where(np.cumsum(pca.explained_variance_)/np.sum(pca.explained_variance_)
                    > 0.90)[0][0]

#    RS=20180824
    tsne=manifold.TSNE(n_components=2, perplexity=perplexity)
    tsne_proj = tsne.fit_transform(pca_proj[:,0:num_comp])
    print(tsne_proj.shape)
#    tsne_proj = tsne.fit_transform(pca_proj[:,0:num_comp])
    tsne_proj_df=pd.DataFrame(index=gene_lists)
    tsne_proj_df["TSNE1"]=tsne_proj[:,0]
    tsne_proj_df["TSNE2"]=tsne_proj[:,1]
    init = tsne_proj_df.loc[marker_genes].values
    num_clusters=len(marker_genes)
    kmeans=KMeans(n_clusters=num_clusters,init = init, random_state=0).fit(tsne_proj)

    tsne_proj_df["cluster"]=kmeans.labels_
    gene_subset_lists=list()
    for geneID in marker_genes:
        gene_subset = tsne_proj_df.index[np.where(tsne_proj_df.cluster == tsne_proj_df.loc[geneID,"cluster"])]
        gene_subset_lists.append(gene_subset)

    for i ,gene_subset in enumerate(gene_subset_lists):
        tsne_proj_df.loc[gene_subset,"cluster"]=i
        
    final_labels = tsne_proj_df.cluster.values
    final_tsne = np.c_[tsne_proj, final_labels]
    palette = sns.color_palette('deep', final_labels.max() + 1)
    colors = [palette[x] if x >= 0 else (0.0, 0.0, 0.0) for x in final_tsne[:,2].astype(int)]
    plt.scatter(final_tsne[:,0], final_tsne[:,1], c=colors, s=28)
    plt.xlabel('TSNE component 1')
    plt.ylabel('TSNE component 2')
    for i in final_labels:
        position = np.max(final_tsne[ final_tsne[:,2]== i], axis=0)
        plt.gcf().gca().text(position[0], position[1]-1,str(i), fontsize=12)
    if fileName != None:
        plt.savefig(fileName,format="pdf",dpi=300)
    plt.show()
    return tsne_proj_df


def plot_cluster_scatter(proj_df,**kw):
    '''
     show cluster scatter
     :param: methods : charater (UMAP or TSNE)
             fileName : charater (savefig path)
             show_labels : bool (True or False to show cluster labels)
    '''
    coor = proj_df.iloc[:,0:2].values
    labels = proj_df.iloc[:,2].values
    palette = sns.color_palette('deep', labels.max() + 1)
    colors = [palette[x] if x >= 0 else (0.0, 0.0, 0.0) for x in labels.astype(int)]
    plt.scatter(coor[:,0],coor[:,1], c=colors, s=28)
    
    methods = kw.get('methods','UMAP')
    plt.xlabel('{} component 1'.format(methods))
    plt.ylabel('{} component 2'.format(methods))
    show_labels = kw.get('show_labels', True)
    if show_labels == True:
        for i in labels:
            position = np.max(coor[labels== i], axis=0)
            plt.gcf().gca().text(position[0], position[1]-1,str(i), fontsize=12)

    fileName = kw.get('fileName', None)
    if fileName !=None:
        plt.savefig(fileName)
    plt.show()

def visualize_cluster_density(proj_df, threshold=0.001, bins=100, ax=None,fig=None,**kw):
    '''
    perform kde density estimationg for tsne/UMAP projection to visualize genes clusters.
    :param file: tsne_proj: shape (m, 2)
    threshold=0.001, bins=100, fileName=None
    '''  
   # fig,ax=plt.subplots()
    coor = proj_df.iloc[:,0:2].values
    kde = gaussian_kde(coor.T, bw_method = 'scott')
    z = kde(coor.T)    
    x = np.ma.masked_where(z > threshold, coor[:,0])
    y = np.ma.masked_where(z > threshold, coor[:,1])
    
    # plot unmasked points
    if proj_df.shape[1]==3:
        labels = proj_df.iloc[:,2].values
        palette = sns.color_palette('deep', labels.max() + 1)
        colors = [palette[x] if x >= 0 else (0.0, 0.0, 0.0) for x in labels.astype(int)]
        ax.scatter(coor[:,0],coor[:,1], c=colors, marker = 'o', s=5)
        for i in labels:
            position = np.max(proj_df[labels== i], axis=0)
            ax.text(position[0], position[1],str(i), fontsize=10)
    else:
        ax.scatter(list(coor[:,0]), list(coor[:,1]), c='black', marker='o', s=5)

    # get bounds from axes
    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()
    
    # prepare grid for density map
    xedges = np.linspace(xmin, xmax, bins)
    yedges = np.linspace(ymin, ymax, bins)
    xx, yy = np.meshgrid(xedges, yedges)
    gridpoints = np.array([xx.ravel(), yy.ravel()])

    # compute density map
    zz = np.reshape(kde(gridpoints), xx.shape)

    # plot density map
    im = ax.imshow(zz, cmap='Spectral_r', interpolation='nearest',
               origin='lower', extent=[xmin, xmax, ymin, ymax],
                  aspect='auto')
    # plot threshold contour
    cs = ax.contour(xx, yy, zz, levels=[threshold], colors='black', line_width=10)
    # show
    fig.colorbar(im,ax=ax)   

    title = kw.get('title', None)
    if 'title' in kw:
        ax.set_title(title,fontsize=12)

    fileName = kw.get('fileName', None)
    if 'fileName' in kw:
        plt.savefig(fileName)

    return z

def dbScan(tsne_proj, z, threshold, eps=1):
    '''
    deprecated
    '''      

    fig, ax = plt.subplots(figsize = (6, 6))
    tsne_proj_sel = tsne_proj[z > threshold]
    db = DBSCAN(eps=eps, min_samples=5).fit(tsne_proj_sel)
    ax.scatter(tsne_proj_sel[:,0], tsne_proj_sel[:,1], c=db.labels_, marker='.')
    n_clusters_ = len(set(db.labels_)) - (1 if -1 in db.labels_ else 0)
    for i in np.arange(n_clusters_):
        position = np.max(tsne_proj_sel[db.labels_ == i], axis=0)
        plt.gcf().gca().text(position[0], position[1]-1,str(i), fontsize=12)
    plt.show()
    return tsne_proj_sel, db.labels_



def spatial_pca_UMAP(data_norm, gene_lists, 
                    n_neighbors = 10, min_dist=0.1):
    '''
    perform standard PCA + UMAP
    :param file: data_norm: normalized gene expression; 
                 gene_lists: list shape(k,)
                 n_neighbors = 10,
                min_dist=0.1
    :rtype: umap_proj: shape (m, 2)
    '''           
    data_s = StandardScaler().fit_transform(data_norm.loc[:, gene_lists])  ## Input matrix (n_sample,n_feature)
    pca = decomposition.PCA()
    pca.fit(data_s.T)
    pca_proj = pca.fit_transform(data_s.T)
    num_comp = np.where(np.cumsum(pca.explained_variance_)/np.sum(pca.explained_variance_) 
                    > 0.9)[0][0]
    umap_proj = umap.UMAP(n_neighbors=n_neighbors,  # 这决定了流形结构局部逼近中相邻点的个数。更大的值将导致更多的全局结构被保留，而失去了详细的局部结构。一般来说，这个参数应该在5到50之间，10到15是一个合理的默认值。
                          min_dist=min_dist,  #这控制了嵌入的紧密程度，允许压缩点在一起。数值越大，嵌入点分布越均匀;数值越小，算法对局部结构的优化越精确。合理的值在0.001到0.5之间，0.1是合理的默认值。
                          n_components=2,
#                           random_state = 0
                         ).fit_transform(pca_proj[:,0:num_comp])  ## genes cluster
    return umap_proj


def spatial_pca_tsne(data_norm, gene_lists, perplexity = 30, random_state=None):
    '''
    perform standard PCA + tsne
    :param file: data_norm: normalized gene expression; gene_lists: list shape(k,)
        perplexity = 30 
    :rtype: tsne_proj: shape (m, 2)
    '''           
    data_s = StandardScaler().fit_transform(data_norm.loc[:, gene_lists])  ## Input matrix (n_sample,n_feature)
    pca = decomposition.PCA()
    pca.fit(data_s.T)
    pca_proj = pca.fit_transform(data_s.T)
    num_comp = np.where(np.cumsum(pca.explained_variance_)/np.sum(pca.explained_variance_) 
                    > 0.9)[0][0]

#    RS=20180824
    tsne=manifold.TSNE(n_components=2,random_state=None, perplexity=perplexity)
    tsne_proj = tsne.fit_transform(pca_proj[:,0:num_comp])
    return tsne_proj




#clustering cells/spots

def create_labels(locs,data_norm,geneList,cellGraph, cluster_size=5,unary_scale_factor=100,smooth_factor=10,rs=0):
    
    
    X=data_norm.loc[:,geneList]
    
    cluster_KM=cluster_size
    kmeans=KMeans(n_clusters=cluster_KM,random_state=rs).fit(X)
    hmrf_labels = cut_graph_profile(cellGraph, kmeans.labels_, unary_scale_factor=unary_scale_factor,
                  smooth_factor=smooth_factor) ## smooth_factor can adjust
    return kmeans.labels_,hmrf_labels

def cut_graph_profile(cellGraph, Kmean_labels, unary_scale_factor=100, 
                      smooth_factor=50, label_cost=10, algorithm='expansion'):
    '''
    Returns new labels and gmm for the cut.
    
    :param points: cellGraph (n,3); count: shape (n,); 
    :unary_scale_factor, scalar; smooth_factor, scalar; 
    :label_cost: scalar; algorithm='expansion'
    :rtype: label shape (n,); gmm object.
    '''

    smooth_factor = smooth_factor
    unary_scale_factor = unary_scale_factor
    label_cost = label_cost
    algorithm = algorithm
    uniq, count = np.unique(Kmean_labels, return_counts = True)  
    unary_cost = compute_unary_cost_profile(Kmean_labels, unary_scale_factor)
    pairwise_cost = compute_pairwise_cost(len(uniq), smooth_factor)
    edges = cellGraph[:,0:2].astype(np.int32)
    labels = pygco.cut_from_graph(edges, unary_cost, pairwise_cost, label_cost)
#    energy = compute_energy(unary_cost, pairwise_cost, edges, labels)

    return labels


def compute_unary_cost_profile(Kmean_labels, scale_factor):
    '''
    Returns unary cost energy.
    
    :param points: count: shape (n,); gmm: gmm object; scale_factor: scalar

    :rtype: unary energy matrix.
    '''    
    labels_pred = Kmean_labels
    uniq, count = np.unique(Kmean_labels, return_counts = True)    
    uninary_mat = np.zeros((len(labels_pred), len(uniq)))
    for i in np.arange(uninary_mat.shape[0]):
        for j in np.arange(len(uniq)):
            if uniq[j] == labels_pred[i]:
                uninary_mat[i, j] = -1
            else:
                uninary_mat[i, j] = 1   
    return (scale_factor*uninary_mat).astype(np.int32)

def count_neighbors(a, b, cellGraph):
    
    idx0 = np.in1d(cellGraph[:,0], a).nonzero()[0]
    idx1 = np.in1d(cellGraph[:,1], a).nonzero()[0]
    neighbor0 = cellGraph[idx0, 1]
    neighbor1 = cellGraph[idx1, 0]
    neighbors = set(neighbor0.tolist() + neighbor1.tolist())   
    
    return len(neighbors.intersection(set(b)))





import numpy as np
import pandas as pd
import matplotlib as mpl
import matplotlib
import matplotlib.cm as cm
import matplotlib.pyplot as plt
import matplotlib.path as mplPath
from scipy.spatial import Voronoi, voronoi_plot_2d, Delaunay, KDTree, ConvexHull
from matplotlib.patches import Polygon
from matplotlib.collections import LineCollection, PatchCollection
from PIL import Image
from matplotlib.backends.backend_pdf import PdfPages
from scipy.stats import gaussian_kde

import seaborn as sns
from scipy.spatial.distance import cdist
from .Graph_cut import *

def plot_cellGraph(locs, cellGraph, 
                scatter_color = 'red', s=1, 
                line_color='blue', linewidth=0.5, 
                title= 'CellGraph', fileName = None):
    fig, ax= plt.subplots(1,1,figsize=(5,5)) #, dpi=300)
    ax.set_aspect('equal')

    ax.scatter(locs[:,0], locs[:,1], s=s, color=scatter_color)
    for i in np.arange(cellGraph.shape[0]):
        x = (locs[int(cellGraph[i,0]), 0], locs[int(cellGraph[i,1]), 0]) 
        y = (locs[int(cellGraph[i,0]), 1], locs[int(cellGraph[i,1]), 1])     
        ax.plot(x, y, color=line_color, linewidth=linewidth)

    plt.title(title)
    plt.show()
    if fileName !=None:
        fig.savefig(fileName)
           

def visualize_spatial_genes(df, locs, data_norm,cellGraph, point_size= 0.5):
    '''
    plot Voronoi tessellation of cells, highlight boundaries of graph cut
    
    :param file: df: dataframe of graph cut results; locs: spatial coordinates (n, 2);
    data_norm: normalized count: shape (n, m); 
    point_size = 0.5; 
    '''    

    i = 0
    while i < df.shape[0]:
        plt.figure(figsize=(6,2.5), dpi=300)
        p1 = plt.subplot(121)
        p2 = plt.subplot(122)

        geneID = df.index[i]
        exp =  data_norm.loc[:,geneID].values
        best_Labels = df.loc[geneID,'label_cell_1':].values.astype(int)
        subplot_voronoi_boundary(geneID, locs, exp, cellGraph, best_Labels,
                                 df.loc[geneID,].fdr, ax=p1, 
                                 fdr=True, point_size = point_size, class_line_width=2)
        i = i + 1
        if i < df.shape[0]:
            geneID = df.index[i]
            exp =  data_norm.loc[:,geneID].values
            best_Labels = df.loc[geneID,'label_cell_1':].values.astype(int)
            subplot_voronoi_boundary(geneID, locs, exp, cellGraph, best_Labels,
                                    df.loc[geneID,].fdr, ax=p2, fdr=True, 
                                     point_size = point_size)    
        else:
            p2.axis('off')
        plt.show()
        i= i + 1


def plot_voronoi_boundary(geneID, coord, count, cellGraph, classLabel, 
                            p, fdr=False, 
                            fileName=None, point_size=5,
                          line_colors="k", class_line_width=2.5,
                          line_width=0.5, line_alpha=1.0,**kw):
    '''
    plot spatial expression as voronoi tessellation
    highlight boundary between classes
    
    :param file: geneID; spatial coordinates shape (n, 2); normalized count: shape (n);
                predicted cell class calls shape (n); prediction p-value.
                fdr=False; line_colors = 'k'; class_line_width = 3; 
                line_width = 0.5; line_alpha = 1.0
    '''
    points = coord
    count = count
    newLabels =classLabel

    # first estimate mean distance between points--
    # p_dist = cdist(points, points)    
    # p_dist[p_dist == 0] = np.max(p_dist, axis = 0)[0]
    # norm_dist = np.mean(np.min(p_dist, axis = 0))
    ## discard p_dist need too much memory to cdist

    norm_dist = np.mean(cellGraph[:,3])     #2021119 coco

    # find points at edge, add three layers of new points 
    x_min = np.min(points, axis = 0)[0] - 3*norm_dist
    y_min = np.min(points, axis = 0)[1] - 3*norm_dist
    x_max = np.max(points, axis = 0)[0] + 3*norm_dist
    y_max = np.max(points, axis = 0)[1] + 3*norm_dist

    n_x = int((x_max - x_min)/norm_dist) + 1
    n_y = int((y_max - y_min)/norm_dist) + 1

    # create a mesh
    x = np.linspace(x_min, x_max, n_x)
    y = np.linspace(y_min, y_max, n_y)
    xv, yv = np.meshgrid(x, y)
    # now select points outside of hull, and merge
    hull = Delaunay(points)
    grid_points = np.hstack((xv.reshape(-1,1), yv.reshape(-1,1)))
    pad_points = grid_points[np.where(hull.find_simplex(grid_points)< 0)[0]]   
    chull = ConvexHull(points)
    outPoints = [points[chull.vertices,0], points[chull.vertices,1]]
    out_coords = np.array([[outPoints[0][vers] ,outPoints[1][vers]] for vers in range(len(outPoints[0]))])
    pad_dist = cdist(pad_points, out_coords) 
    # pad_dist = cdist(pad_points, points)   ## discard func, need too much memory to cdist
    pad_points = pad_points[np.where(np.min(pad_dist, axis = 1) > norm_dist)[0]]
    all_points = np.vstack((points, pad_points))

    ori_len = points.shape[0]
    vor = Voronoi(all_points)

    if kw.get("show_points",True):
        plt.plot(points[0:ori_len,0], points[0:ori_len,1], ".", markersize=point_size)     
    patches = []
    # but we onl use the original points fot plotting
    for i in np.arange(ori_len):
        good_ver = vor.vertices[vor.regions[vor.point_region[i]]]
        polygon = Polygon(good_ver, True)
        patches.append(polygon)

    pc = PatchCollection(patches, cmap= cm.PiYG, alpha=1)

    pc.set_array(np.array(count))

    plt.gca().add_collection(pc)
    # for loop for plotting is slow, consider to vectorize to speedup
    # doesn;t mater for now unless you have many point or genes
    finite_segments=[]
    boundary_segments=[]
    for kk, ii in vor.ridge_dict.items():
        if kk[0] < ori_len and kk[1] < ori_len:
            if newLabels[kk[0]] !=  newLabels[kk[1]]:
                boundary_segments.append(vor.vertices[ii])
            else:
                finite_segments.append(vor.vertices[ii])

    plt.gca().add_collection(LineCollection(boundary_segments,
                                            colors="k",           
                                            lw=class_line_width,
                                            alpha=1,
                                            linestyles="solid"))
    plt.gca().add_collection(LineCollection(finite_segments,
                                            colors=line_colors,
                                            lw=line_width,
                                            alpha=line_alpha,
                                            linestyle="solid"))

    plt.xlim(x_min + 1*norm_dist, x_max - 1*norm_dist)
    plt.ylim(y_min + 1*norm_dist, y_max - 1*norm_dist)

    # also remember to add color bar
    plt.colorbar(pc)
    
    if fdr:
            titleText = geneID + '\n' + 'fdr: ' + str("{:.2e}".format(p))
    else:
            titleText = geneID + '\n' + 'p_value: ' + str("{:.2e}".format(p))
    
    titleText=kw.get("set_title",titleText)
    fontsize=kw.get("fontsize",12)
    plt.title(titleText, fontname="Arial", fontsize=fontsize)
    plt.xlabel('X coordinate')
    plt.ylabel('Y coordinate')
    if fileName!=None:
        plt.savefig(fileName)
    plt.show() 



def pdf_voronoi_boundary(geneID, coord, count, cellGraph, classLabel,
                         p , fileName, fdr=False,  point_size=5,
                          line_colors="k", class_line_width=2.5,
                          line_width=0.5, line_alpha=1.0,**kw):
    '''
    save spatial expression as voronoi tessellation to pdf
    highlight boundary between classes.
    
    :param file: geneID; spatial coordinates shape (n, 2); normalized count: shape (n);
                predicted cell class calls shape (n); prediction p-value; pdf fileName;
                fdr=False; line_colors = 'k'; class_line_width = 3; 
                line_width = 0.5; line_alpha = 1.0
    '''

    points = coord
    count = count
    newLabels =classLabel

    # first estimate mean distance between points--
    # p_dist = cdist(points, points)    
    # p_dist[p_dist == 0] = np.max(p_dist, axis = 0)[0]
    # norm_dist = np.mean(np.min(p_dist, axis = 0))
     ## discard p_dist need too much memory to cdist

    norm_dist = np.mean(cellGraph[:,3])

    # find points at edge, add three layers of new points 
    x_min = np.min(points, axis = 0)[0] - 3*norm_dist
    y_min = np.min(points, axis = 0)[1] - 3*norm_dist
    x_max = np.max(points, axis = 0)[0] + 3*norm_dist
    y_max = np.max(points, axis = 0)[1] + 3*norm_dist

    n_x = int((x_max - x_min)/norm_dist) + 1
    n_y = int((y_max - y_min)/norm_dist) + 1

    # create a mesh
    x = np.linspace(x_min, x_max, n_x)
    y = np.linspace(y_min, y_max, n_y)
    xv, yv = np.meshgrid(x, y)
    # now select points outside of hull, and merge
    hull = Delaunay(points)
    grid_points = np.hstack((xv.reshape(-1,1), yv.reshape(-1,1)))
    pad_points = grid_points[np.where(hull.find_simplex(grid_points)< 0)[0]]
    chull = ConvexHull(points) # convex hull points
    outPoints = [points[chull.vertices,0], points[chull.vertices,1]]
    out_coords = np.array([[outPoints[0][vers] ,outPoints[1][vers]] for vers in range(len(outPoints[0]))])
    pad_dist = cdist(pad_points, out_coords)
    # pad_dist = cdist(pad_points, points)   
    pad_points = pad_points[np.where(np.min(pad_dist, axis = 1) > norm_dist)[0]]
    all_points = np.vstack((points, pad_points))

    ori_len = points.shape[0]
    vor = Voronoi(all_points)

    if kw.get("show_points",True):
        plt.plot(points[0:ori_len,0], points[0:ori_len,1], ".", markersize=point_size)     
    patches = []
    # but we onl use the original points fot plotting
    for i in np.arange(ori_len):
        good_ver = vor.vertices[vor.regions[vor.point_region[i]]]
        polygon = Polygon(good_ver, True)
        patches.append(polygon)

    pc = PatchCollection(patches, cmap=cm.PiYG, alpha=1)

    pc.set_array(np.array(count))

    plt.gca().add_collection(pc)
    # for loop for plotting is slow, consider to vectorize to speedup
    # doesn;t mater for now unless you have many point or genes
    finite_segments=[]
    boundary_segments=[]
    for kk, ii in vor.ridge_dict.items():
        if kk[0] < ori_len and kk[1] < ori_len:
            if newLabels[kk[0]] !=  newLabels[kk[1]]:
                boundary_segments.append(vor.vertices[ii])
            else:
                finite_segments.append(vor.vertices[ii])

    plt.gca().add_collection(LineCollection(boundary_segments,
                                            colors="k",           
                                            lw=class_line_width,
                                            alpha=1,
                                            linestyles="solid"))
    plt.gca().add_collection(LineCollection(finite_segments,
                                            colors=line_colors,
                                            lw=line_width,
                                            alpha=line_alpha,
                                            linestyle="solid"))
    plt.xlim(x_min + 1*norm_dist, x_max - 1*norm_dist)
    plt.ylim(y_min + 1*norm_dist, y_max - 1*norm_dist)

    # also remember to add color bar
    plt.colorbar(pc)
    
    if fdr:
            titleText = geneID + '\n' + 'fdr: ' + str("{:.2e}".format(p))
    else:
            titleText = geneID + '\n' + 'p_value: ' + str("{:.2e}".format(p))
    
    titleText=kw.get("set_title",titleText)
    fontsize=kw.get("fontsize",12)
    plt.title(titleText, fontname="Arial", fontsize=fontsize)
    plt.axis('off')
#    plt.xlabel('X coordinate')
#    plt.ylabel('Y coordinate')
    if fileName != None:
        plt.savefig(fileName)
    else:
        print('ERROR! Please supply a file name.')



def subplot_voronoi_boundary(geneID, coord, count, cellGraph, classLabel, 
                            p  ,ax ,
                            fdr=False, point_size=5,
                          line_colors="k", class_line_width=2.5,
                          line_width=0.5, line_alpha=1.0,**kw):
    '''
    plot spatial expression as voronoi tessellation
    highlight boundary between classes
    
    :param file: geneID; spatial coordinates (n, 2); normalized gene expression: count;
            predicted cell class calls (n); p_value; ax number;
    '''
    points = coord
    count = count
    newLabels =classLabel

    # first estimate mean distance between points--
    # p_dist = cdist(points, points)    
    # p_dist[p_dist == 0] = np.max(p_dist, axis = 0)[0]
    # norm_dist = np.mean(np.min(p_dist, axis = 0))
    ## discard p_dist need too much memory to cdist

    norm_dist = np.mean(cellGraph[:,3])

    # find points at edge, add three layers of new points 
    x_min = np.min(points, axis = 0)[0] - 3*norm_dist
    y_min = np.min(points, axis = 0)[1] - 3*norm_dist
    x_max = np.max(points, axis = 0)[0] + 3*norm_dist
    y_max = np.max(points, axis = 0)[1] + 3*norm_dist

    n_x = int((x_max - x_min)/norm_dist) + 1
    n_y = int((y_max - y_min)/norm_dist) + 1

    # create a mesh
    x = np.linspace(x_min, x_max, n_x)
    y = np.linspace(y_min, y_max, n_y)
    xv, yv = np.meshgrid(x, y)
    # now select points outside of hull, and merge
    hull = Delaunay(points)
    grid_points = np.hstack((xv.reshape(-1,1), yv.reshape(-1,1)))
    pad_points = grid_points[np.where(hull.find_simplex(grid_points)< 0)[0]]
    chull = ConvexHull(points) # convex hull points
    outPoints = [points[chull.vertices,0], points[chull.vertices,1]]
    out_coords = np.array([[outPoints[0][vers] ,outPoints[1][vers]] for vers in range(len(outPoints[0]))])
    pad_dist = cdist(pad_points, out_coords)
    # pad_dist = cdist(pad_points, points)   
    pad_points = pad_points[np.where(np.min(pad_dist, axis = 1) > norm_dist)[0]]
    all_points = np.vstack((points, pad_points))

    ori_len = points.shape[0]
    vor = Voronoi(all_points)

    if kw.get("show_points",True):
        ax.plot(points[0:ori_len,0], points[0:ori_len,1], ".", markersize=point_size)  

    ## plt.full(color)   
    patches = []
    # but we onl use the original points fot plotting
    for i in np.arange(ori_len):
        good_ver = vor.vertices[vor.regions[vor.point_region[i]]]
        polygon = Polygon(good_ver, True)
        patches.append(polygon)

    pc = PatchCollection(patches, cmap=cm.PiYG, alpha=1)

    pc.set_array(np.array(count))

    ax.add_collection(pc)


    # for loop for plotting is slow, consider to vectorize to speedup
    # doesn;t mater for now unless you have many point or genes
    finite_segments=[]
    boundary_segments=[]
    for kk, ii in vor.ridge_dict.items():
        if kk[0] < ori_len and kk[1] < ori_len:
            if newLabels[kk[0]] !=  newLabels[kk[1]]:
                boundary_segments.append(vor.vertices[ii])
            else:
                finite_segments.append(vor.vertices[ii])

    ax.add_collection(LineCollection(boundary_segments,
                                            colors="k",           
                                            lw=class_line_width,
                                            alpha=1,
                                            linestyles="solid"))
    ax.add_collection(LineCollection(finite_segments,
                                            colors=line_colors,
                                            lw=line_width,
                                            alpha=line_alpha,
                                            linestyle="solid"))
    ax.set_xlim(x_min + 1*norm_dist, x_max - 1*norm_dist)
    ax.set_ylim(y_min + 1*norm_dist, y_max - 1*norm_dist)

    # also remember to add color bar
    #plt.colorbar(pc)
    

    if fdr:
            titleText = geneID + '\n' + 'fdr: ' + str("{:.2e}".format(p))
    else:
            titleText = geneID + '\n' + 'p_value: ' + str("{:.2e}".format(p))
    
    titleText=kw.get("set_title",titleText)
    fontsize=kw.get("fontsize",8)
    ax.set_title(titleText, fontname="Arial", fontsize=fontsize)



def subplot_voronoi_boundary_12x18(geneID, coord, count, cellGraph,
                          classLabel, p, ax, fdr=False, point_size = 0.5,  
                          line_colors = 'k', class_line_width = 0.8, 
                          line_width = 0.05, line_alpha = 1.0,**kw):
    '''
    plot spatial expression as voronoi tessellation
    highlight boundary between classes
    
    :param file: geneID; coord: spatial coordinates (n, 2); count: normalized gene expression;
        predicted cell class calls (n); p: graph cut p-value. 
    '''
    points = coord
    count = count
    newLabels =classLabel

    # p_dist = cdist(points, points)    
    # p_dist[p_dist == 0] = np.max(p_dist, axis = 0)[0]
    # norm_dist = np.mean(np.min(p_dist, axis = 0))
    ## discard p_dist need too much memory to cdist

    norm_dist = np.mean(cellGraph[:,3])

    # find points at edge, add three layers of new points 
    x_min = np.min(points, axis = 0)[0] - 3*norm_dist
    y_min = np.min(points, axis = 0)[1] - 3*norm_dist
    x_max = np.max(points, axis = 0)[0] + 3*norm_dist
    y_max = np.max(points, axis = 0)[1] + 3*norm_dist

    n_x = int((x_max - x_min)/norm_dist) + 1
    n_y = int((y_max - y_min)/norm_dist) + 1

    # create a mesh
    x = np.linspace(x_min, x_max, n_x)
    y = np.linspace(y_min, y_max, n_y)
    xv, yv = np.meshgrid(x, y)
    # now select points outside of hull, and merge
    hull = Delaunay(points)
    grid_points = np.hstack((xv.reshape(-1,1), yv.reshape(-1,1)))
    pad_points = grid_points[np.where(hull.find_simplex(grid_points)< 0)[0]]
    chull = ConvexHull(points) # convex hull points
    outPoints = [points[chull.vertices,0], points[chull.vertices,1]]
    out_coords = np.array([[outPoints[0][vers] ,outPoints[1][vers]] for vers in range(len(outPoints[0]))])
    pad_dist = cdist(pad_points, out_coords)
    # pad_dist = cdist(pad_points, points)   
    pad_points = pad_points[np.where(np.min(pad_dist, axis = 1) > norm_dist)[0]]
    all_points = np.vstack((points, pad_points))

    ori_len = points.shape[0]
    vor = Voronoi(all_points)
    if kw.get("show_points",True):
        ax.plot(points[0:ori_len,0], points[0:ori_len,1], ".", markersize=point_size)     
    patches = []
    # but we onl use the original points fot plotting
    for i in np.arange(ori_len):
        good_ver = vor.vertices[vor.regions[vor.point_region[i]]]
        polygon = Polygon(good_ver, True)
        patches.append(polygon)

    pc = PatchCollection(patches, cmap=cm.PiYG, alpha=1)

    pc.set_array(np.array(count))

    ax.add_collection(pc)
    # for loop for plotting is slow, consider to vectorize to speedup
    # doesn;t mater for now unless you have many point or genes
    finite_segments=[]
    boundary_segments=[]
    for kk, ii in vor.ridge_dict.items():
        if kk[0] < ori_len and kk[1] < ori_len:
            if newLabels[kk[0]] !=  newLabels[kk[1]]:
                boundary_segments.append(vor.vertices[ii])
            else:
                finite_segments.append(vor.vertices[ii])

    ax.add_collection(LineCollection(boundary_segments,   ### boundary
                                            colors="k",           
                                            lw=class_line_width,
                                            alpha=1,
                                            linestyles="solid"))
    ax.add_collection(LineCollection(finite_segments,               ## other line in loop
                                            colors=line_colors,
                                            lw=line_width,
                                            alpha=line_alpha,
                                            linestyle="solid"))
    ax.set_xlim(x_min + 1*norm_dist, x_max - 1*norm_dist)
    ax.set_ylim(y_min + 1*norm_dist, y_max - 1*norm_dist)

    # also remember to add color bar
    #plt.colorbar(pc)

    if fdr:
        titleText = geneID + ' ' + '' + str("{0:.1e}".format(p))
    else:
        titleText = geneID + ' ' + 'p_value: ' + str("{0:1e}".format(p))
    titleText=kw.get("set_title",titleText)
    fontsize=kw.get("fontsize",3.5)
    y = kw.get('set_y',0.85)
    ax.set_title(titleText, fontname="Arial", fontsize=fontsize, y = y)



def multipage_pdf_visualize_spatial_genes(df, locs, data_norm, cellGraph,
                                          fileName, point_size=0,
                                          line_colors = 'k', 
                                          class_line_width = 0.8, 
                                          line_width = 0.05, 
                                            line_alpha = 1.0,**kw):
    
#     '''
#     save spatial expression as voronoi tessellation to pdf highlight boundary between classes
#     format: 12 by 18.
#     :param file: df: graph cuts results; locs: spatial coordinates (n, 2); data_norm: normalized gene expression;
#         pdf filename; point_size=0.5. 
#     '''    
 
    points = locs
    vor = Voronoi(points)

    nb_plots = int(df.shape[0])
    numCols = kw.get('ncols', 12)  
    numRows = kw.get('nrows', 18)
    nb_plots_per_page =  numCols*numRows
    t_numRows = int(df.shape[0]/numCols) + 1
    
    with PdfPages(fileName) as pdf:    
        for i in np.arange(df.shape[0]):
            if i % nb_plots_per_page == 0:
                fig, axs = plt.subplots(numRows, numCols, # 8 11
                                    figsize = (8,11))   
                fig.subplots_adjust(hspace=0.3, wspace=0.3,
                                top=0.925, right=0.925, bottom=0.075, left = 0.075)
                  
            geneID = df.index[i]
            exp =  data_norm.loc[:,geneID].values
            p = df.loc[geneID,].fdr

            dist_df = kw.get('dist_df',None)
            if dist_df != None:
                ham = dist_df.loc[geneID,'Hamming']
                jac = dist_df.loc[geneID, 'Jaccard']
                haus = dist_df.loc[geneID, 'Hausdorff']
#                 diff_Vs_common = dist_df.loc[geneID,'Hamming2']
                titleText = geneID + '\n'+'Ham:{:.2f}; Jac:{:.2f}'.format(ham, jac)+'\n'+'Hausdorff:{:.2f}'.format(haus)
                y = 0.75
            else:
                titleText = geneID + ' ' + '' + str("{0:.1e}".format(p))
                y = 0.85

            if np.isnan(df.loc[geneID,].fdr):
                best_Labels = np.zeros(data_norm.shape[0])
            else:
                best_Labels = df.loc[geneID,'label_cell_1':].values.astype(int)
            m = int(i/numCols) % numRows
            n = i % numCols 
            ax = axs[m,n]
            subplot_voronoi_boundary_12x18(geneID, locs, exp, cellGraph, 
                                best_Labels,
                                p, ax=ax, fdr=True,
                                point_size = point_size,  
                                line_colors = line_colors, 
                                class_line_width = class_line_width, 
                                line_width = line_width, 
                                line_alpha = line_alpha,
                                set_title=titleText,
                                set_y= y)

            if (i + 1) % nb_plots_per_page == 0 or (i + 1) == nb_plots:
                for ii in np.arange(numRows):
                    for jj in np.arange(numCols):        
                        axs[ii,jj].axis('off')
                pdf.savefig(fig)
                fig.clear()
                plt.close()
            


def add_HE_image(image,ax):
    img=Image.open(image)
    extent_size = [1,33,1,35]
    img_transpose=img.transpose(Image.FLIP_TOP_BOTTOM)
    ax.imshow(img_transpose,extent=extent_size)


def subplot_boundary(geneID, coord, count, cellGraph,classLabel, 
                    p, ax=None,
                    fdr=False, point_size=5,
                    class_line_width=2.5,
                    **kw):
    
    '''
    plot spatial expression as voronoi tessellation.
    :param file: geneID; spatial coordinates (n, 2); normalized count: shape (n); 
    '''
    
    points = coord
    count = count
    newLabels =classLabel

    # first estimate mean distance between points--
    # p_dist = cdist(points, points)    
    # p_dist[p_dist == 0] = np.max(p_dist, axis = 0)[0]
    # norm_dist = np.mean(np.min(p_dist, axis = 0))
    ## discard p_dist need too much memory to cdist

    norm_dist = np.mean(cellGraph[:,3])

    # find points at edge, add three layers of new points 
    x_min = np.min(points, axis = 0)[0] - 3*norm_dist
    y_min = np.min(points, axis = 0)[1] - 3*norm_dist
    x_max = np.max(points, axis = 0)[0] + 3*norm_dist
    y_max = np.max(points, axis = 0)[1] + 3*norm_dist

    n_x = int((x_max - x_min)/norm_dist) + 1
    n_y = int((y_max - y_min)/norm_dist) + 1

    # create a mesh
    x = np.linspace(x_min, x_max, n_x)
    y = np.linspace(y_min, y_max, n_y)
    xv, yv = np.meshgrid(x, y)
    # now select points outside of hull, and merge
    hull = Delaunay(points)
    grid_points = np.hstack((xv.reshape(-1,1), yv.reshape(-1,1)))
    pad_points = grid_points[np.where(hull.find_simplex(grid_points)< 0)[0]]
    chull = ConvexHull(points) # convex hull points
    outPoints = [points[chull.vertices,0], points[chull.vertices,1]]
    out_coords = np.array([[outPoints[0][vers] ,outPoints[1][vers]] for vers in range(len(outPoints[0]))])
    pad_dist = cdist(pad_points, out_coords)
    # pad_dist = cdist(pad_points, points)   
    pad_points = pad_points[np.where(np.min(pad_dist, axis = 1) > norm_dist)[0]]
    all_points = np.vstack((points, pad_points))

    ori_len = points.shape[0]
    vor = Voronoi(all_points)

    if kw.get("show_points",False):
        ax.plot(points[0:ori_len,0], points[0:ori_len,1], ".", markersize=point_size)
    
    # for loop for plotting is slow, consider to vectorize to speedup
    # doesn;t mater for now unless you have many point or genes
    boundary_segments=[]
    for kk, ii in vor.ridge_dict.items():
        if kk[0] < ori_len and kk[1] < ori_len:
            if newLabels[kk[0]] !=  newLabels[kk[1]]:
                boundary_segments.append(vor.vertices[ii])

    ax.add_collection(LineCollection(boundary_segments,
                                            colors="k",           
                                            lw=class_line_width,
                                            alpha=1,
                                            linestyles="solid"))
    
    ax.set_xlim(x_min + 1*norm_dist, x_max - 1*norm_dist)
    ax.set_ylim(y_min + 1*norm_dist, y_max - 1*norm_dist)
    
    if fdr:
            titleText = geneID + '\n' + 'fdr: ' + str("{:.2e}".format(p))
    else:
            titleText = geneID + '\n' + 'p_value: ' + str("{:.2e}".format(p))
    
    titleText=kw.get("set_title",titleText)
    fontsize=kw.get("fontsize",8)
    ax.set_title(titleText, fontname="Arial", fontsize=8)


def plot_tissue_pattern(locs,data_norm, cellGraph, tissue_mat,image,colors,title,nrows=4,ncols=5,s=15):
        ## Task2: Tissue mat
    nb_plots=tissue_mat.shape[0]
    nrows=nrows
    ncols=ncols
    nb_box=nrows*ncols
  
    fig,ax=plt.subplots(nrows,ncols,figsize=(ncols*3,nrows*3),dpi=180)
    fig.subplots_adjust(hspace=0.3, wspace=0.3,
                                    top=0.925, right=0.925, bottom=0.075, left = 0.075)
    
    for i in range(tissue_mat.shape[0]):
        x=int(i/ncols)
        y=i%ncols
        axes=ax[x,y]
        
        add_HE_image(image,axes)
        axes.scatter(locs[:,0], locs[:,1], c=tissue_mat[i],
                     cmap=matplotlib.colors.ListedColormap(colors) ,s=s)
        
        axes.set_title(title,fontsize=8)


        points=locs
        # p_dist = cdist(points, points)    
        # p_dist[p_dist == 0] = np.max(p_dist, axis = 0)[0]
        # norm_dist = np.mean(np.min(p_dist, axis = 0))
        ## discard p_dist need too much memory to cdist
        norm_dist = np.mean(cellGraph[:,3])

        # find points at edge, add three layers of new points 
        x_min = np.min(points, axis = 0)[0] - 3*norm_dist
        y_min = np.min(points, axis = 0)[1] - 3*norm_dist
        x_max = np.max(points, axis = 0)[0] + 3*norm_dist
        y_max = np.max(points, axis = 0)[1] + 3*norm_dist

        axes.set_xlim(x_min + 1*norm_dist, x_max - 1*norm_dist)
        axes.set_ylim(y_min + 1*norm_dist, y_max - 1*norm_dist)

        if (i + 1) == nb_plots:
            for ii in np.arange(nb_plots,nb_box):        
                    ax[int(ii/ncols),ii%ncols].axis('off')

def subplot_HE_with_labels(locs,cellGraph, labels,image,ax,colors,title,s=30):
    
   # import matplotlib
    add_HE_image(image,ax)  
    ax.scatter(locs[:,0], locs[:,1], c=labels,
                 cmap=matplotlib.colors.ListedColormap(colors) ,s=s)
    
    ax.set_title(title,fontsize=8)

    points=locs
    # p_dist = cdist(points, points)    
    # p_dist[p_dist == 0] = np.max(p_dist, axis = 0)[0]
    # norm_dist = np.mean(np.min(p_dist, axis = 0))
    ## discard p_dist need too much memory to cdist
    norm_dist = np.mean(cellGraph[:,3])
    x_min = np.min(points, axis = 0)[0] - 3*norm_dist
    y_min = np.min(points, axis = 0)[1] - 3*norm_dist
    x_max = np.max(points, axis = 0)[0] + 3*norm_dist
    y_max = np.max(points, axis = 0)[1] + 3*norm_dist

    ax.set_xlim(x_min + 1*norm_dist, x_max - 1*norm_dist)
    ax.set_ylim(y_min + 1*norm_dist, y_max - 1*norm_dist)


def plot_multi_bars(datas, labels, group_name,colors, tick_step = 1, group_gap= 0.2, bar_gap=0, **options):
    
    """
    para: datas: multi groups datasets, 2-D list, row is methods, col is group.
        labels: methods names
        group_name: group names
        tick_step : x axis stride
        group_gap: the width of between group 
        bar_gap: the width of each bars
    """
    # ticks is x_axis ticks
    x_ticks = np.arange(len(group_name))
    ticks =  x_ticks*tick_step
    group_num = len(datas)
    
    #
    group_width = tick_step - group_gap
    bar_span = group_width/group_num
    bar_width = bar_span - bar_gap
    
    alpha = options.get('alpha', 0.7)
    rotation = options.get('rotation',30)
    loc = options.get('loc','upper right')
    
    baseline_x = ticks - (group_width - bar_span)/2
    for index, y in enumerate(datas):
        plt.bar(baseline_x+index*bar_span, y, bar_width,
                label = labels[index],
               color = colors[index],
               alpha = alpha)
    plt.xticks(x_ticks, group_name, rotation = rotation)
    plt.legend(loc=loc,frameon=False)
        


from setuptools import setup, find_packages
setup(
        name='scGCO',
        version='1.1.2',
        description='single-cell graph cuts optimization',
        url='https://github.com/WangPeng-Lab/scGCO',
        packages=find_packages(),
        include_package_data=True,
        install_requires=['pandas','numpy','matplotlib','scipy','sklearn','seaborn','parmap','Cython','pygco',
                        'tqdm','networkx','shapely','statsmodels','hdbscan','pillow','scikit-image','umap-learn','pysal==2.0.0'],
        author='Peng Wang',
        author_email='wangpeng@pb.ac.cn',
        license='MIT'

)

## pip install numpy
## pip install pygco
## conda install -c conda-forge cython
## pip uninstall shapely && conda install -c r shapely   



from .Preprocessing import *
from .Graph_cut import *
from .Test_FPR import *
from .Clustering import *
from .Visualization import *
from .Calculate_dist import *
from .venn import *
from .GMM import *